import skimage.io as io
from skoots.lib.skeleton import skeleton_to_mask
from skoots.train.merged_transform import _get_affine_matrix, _get_inverse_affine_matrix, transform_from_cfg
import torch
from torch import Tensor
from torchvision.transforms.functional import affine
import matplotlib.pyplot as plt
from skimage.morphology import skeletonize
from skoots.config import get_cfg_defaults
import numpy as np

image = io.imread("/home/chris/Dropbox (Partners HealthCare)/skoots/tests/hide_test_2d.tif")
skeletons = {}
for u in np.unique(image):
    if u == 0: continue
    skel = skeletonize(image == u)
    skel = torch.from_numpy(skel).unsqueeze(-1).nonzero().round().float()
    skeletons[u] = skel


# x = torch.linspace(50, 200, 10).unsqueeze(1)
# y = torch.ones((10,1)) * 250/2
# z = torch.zeros((10,1))
#
# s = torch.concat((x,y,z), dim=1)

# unrotated_skeleton_mask = skeleton_to_mask(skeletons, (image.shape[0], image.shape[1], 1))

image = torch.from_numpy(image.astype(float)).unsqueeze(0).unsqueeze(-1)

# angle = torch.tensor((0))
# scale = torch.tensor((1))
# shear = torch.tensor((15))
# for s in range(1):
#     shear = torch.tensor((s)) + 15
#     mat: Tensor = _get_affine_matrix(
#         center=[image.shape[1] / 2, image.shape[2] / 2],
#         angle=angle.item(),
#         translate=[0.0, 0.0],
#         scale=scale.item(),
#         shear=[0.0, float(shear.item())],
#         device=str(image.device),
#     )
#
#
#     # print(mat / image.shape[1] / 2)
#     # _mat = _get_inverse_affine_matrix(
#     #     center=[image.shape[1] / 2, image.shape[2] / 2],
#     #     angle=-angle.item(),
#     #     translate=[0.0, 0.0],
#     #     scale=scale.item(),
#     #     shear=[float(shear.item()), float(shear.item())], #float(shear.item())],
#     #     inverted=False,
#     # )
#     # _mat += [0, 0, 1]
#     # _mat = torch.tensor(_mat).view(3,3)
#
#
#     # Rotate the skeletons by the affine matrix
#     for k, v in skeletons.items():
#         skeleton_xy = v[:, [0, 1]].permute(1, 0).unsqueeze(0)  # [N, 3] -> [1, 2, N]
#         _ones = torch.ones(
#             (1, 1, skeleton_xy.shape[-1]), device='cpu'
#         )  # [1, 1, N]
#         skeleton_xy = torch.cat((skeleton_xy, _ones), dim=1)  # [1, 3, N]
#         rotated_skeleton = mat @ skeleton_xy  # [1,3,N]
#         skeletons[k][:, [0, 1]] = rotated_skeleton[0, [0, 1], :].T.float()
#
#
#     rotated = skeleton_to_mask(skeletons, (image.shape[1], image.shape[2], 1))
#     rotated_image = affine(
#         image.permute(0, 3, 1, 2).float(),
#         angle=angle.item(),
#         shear=float(shear.item()),
#         scale=scale.item(),
#         translate=[0, 0],
#     ).permute(0, 2, 3, 1)

    # rotated_image[rotated > 0.5] = rotated_image.max() + 2
    # plt.imshow(rotated_image.squeeze().numpy())
    # plt.title('before')
    # plt.show()
    # plt.close()

def load_cfg_from_file(path):
    """Load configurations."""
    # Set configurations
    cfg = get_cfg_defaults()
    cfg.merge_from_file(path)
    cfg.freeze()
    return cfg


cfg = load_cfg_from_file('/home/chris/Dropbox (Partners HealthCare)/skoots-experiments/configs/mitochondria/test_fine_tune_skoots_on_all.yaml')

skeletons = {k:v.cuda() for k,v in skeletons.items()}

dd = {'image': image.cuda(), 'masks': image.cuda(), 'skeletons': skeletons}
for _ in range(1):
    out = transform_from_cfg(dd, cfg)
    rotated = skeleton_to_mask(out['skeletons'], (out['image'].shape[1], out['image'].shape[2], 1))
    out['image'][rotated > 0.5] = out['image'].max() + 2
    plt.imshow(out['image'].squeeze().cpu().numpy())
    plt.show()
    plt.close()





from typing import List, Tuple, Dict

import torch
import torch.nn.functional as F
import triton
import triton.language as tl
from skoots.lib.morphology import (
    gauss_filter,
    binary_dilation_2d,
    _compute_zero_padding,
    _get_binary_kernel3d,
)
from skoots.lib.utils import get_cached_ball_coords, get_cached_disk_coords
from torch import Tensor
import logging


@torch.jit.script
def average_baked_skeletons(baked_skeleton: Tensor, kernel_size: int = 3) -> Tensor:
    """
    Takes a baked skeleton computed by skoots.lib.skeleton.bake_skeleton and averages all the values such that
    there is a smooth transition from one location to another.

    Shapes:
        - baked_skeleton :math:`(B_{in}, 3, X_{in}, Y_{in}, Z_{in})`
        - returns :math:`(B_{in}, 3, X_{in}, Y_{in}, Z_{in})`

    :param baked_skeleton: Baked skeleton tensor
    :param kernel_size: Kernel size for smoothing
    :return: smoothed skeleton: Smoothed Tensor
    """
    padding: Tuple[int, int, int] = _compute_zero_padding(
        (kernel_size, kernel_size, kernel_size)
    )
    kernel: Tensor = _get_binary_kernel3d(kernel_size, str(baked_skeleton.device))
    b, c, h, w, d = baked_skeleton.shape

    # map the local window to single vector
    features: Tensor = F.conv3d(
        baked_skeleton.reshape(b * c, 1, h, w, d), kernel, padding=padding, stride=1
    )
    features: Tensor = features.view(b, c, -1, h, w, d)  # B, C, -1, X, Y, Z

    nonzero = features.gt(0).sum(2)  # B, C, X, Y, Z
    nonzero[nonzero.eq(0)] = 1.0
    features = features.sum(
        2
    )  # This is the average of the kernel window... without zeros
    features = features / nonzero

    return features


@triton.jit
def _min_skeleton_kernel(
    # Pointers to mat
    mask_pointer,
    skeleton_pointer,
    skeleton_id_map_pointer,
    skeleton_len_pointer,
    out_pointer,
    anisotopy_pointer,  # 1D tensor
    # shapes of the MASK/OUT
    mask_x_shape,
    mask_y_shape,
    mask_z_shape,
    out_c_shape,  # should be 3
    # strides of the MASK/OUT
    anisotropy_stride,
    mask_x_stride,
    mask_y_stride,
    mask_z_stride,
    out_c_stride,  # should be the x,y,z channel of out
    out_x_stride,
    out_y_stride,
    out_z_stride,
    # skeleton shape
    n_skeleton: tl.constexpr,
    dim_skeleton,  # should always be three
    # skeleton strides
    # id_skeleton_stride, # skeleton will be a [N_objects, M_points, C=3] Tensor
    n_skeleton_stride,
    m_skeleton_stride,
    dim_skeleton_stride,

    SKEL_BLOCK_SIZE: tl.constexpr,
    ID_BLOCK_SIZE: tl.constexpr,
):
    """
    This is a triton kernel for calculating the closest skeleton vertex of a to a voxel in an instance mask, where each
    ID has a unique skeleton. Each instance mask voxel is presumed to contain a 0 (background) or integer id,
    signifying this voxel belongs to the object with that id.

    The mask is a 3D tensor of shape (X, Y, Z) with N objects in it. Each N object has a unique ID, but is not
    required to be sequential. I.e. there can be two IDs in mask: 234 and 6.

    Each objects skeleton consists of 3D points; with the number of points varrying between objects.
    This kernel combines

    :param mask_pointer:
    :param skeleton_pointer:
    :param skeleton_id_map_pointer:
    :param skeleton_len_pointer:
    :param out_pointer:
    :param anisotopy_pointer:
    :param mask_x_shape:
    :param mask_y_shape:
    :param mask_z_shape:
    :param out_c_shape:
    :param anisotropy_stride:
    :param mask_x_stride:
    :param mask_y_stride:
    :param mask_z_stride:
    :param out_c_stride:
    :param out_x_stride:
    :param out_y_stride:
    :param out_z_stride:
    :param n_skeleton:
    :param dim_skeleton:
    :param n_skeleton_stride:
    :param m_skeleton_stride:
    :param dim_skeleton_stride:
    :param SKEL_BLOCK_SIZE:
    :param ID_BLOCK_SIZE:
    :return:
    """

    # padding

    x0 = tl.program_id(axis=0)
    y0 = tl.program_id(axis=1)
    z0 = tl.program_id(axis=2)

    mask_center = tl.load(
        mask_pointer + (x0 * mask_x_stride + y0 * mask_y_stride + z0 * mask_z_stride)
    )

    if (
        mask_center == 0.0
    ):  # depending on how triton does this, it might be scuffed
        return

    dx = tl.load(anisotopy_pointer + (0 * anisotropy_stride))  # this should always be 3
    dy = tl.load(anisotopy_pointer + (1 * anisotropy_stride))  # this should always be 3
    dz = tl.load(anisotopy_pointer + (2 * anisotropy_stride))  # this should always be 3

    _off = tl.arange(0, ID_BLOCK_SIZE)  # assume stride is one here. Verified elsewhere
    skeleton_id_map = tl.load(skeleton_id_map_pointer + _off, mask=_off < n_skeleton)
    skeleton_len_map = tl.load(skeleton_len_pointer + _off, mask=_off < n_skeleton)

    tl.device_assert(tl.sum(skeleton_len_map > 0) == n_skeleton, 'it is broken')


    index = tl.argmax(skeleton_id_map == mask_center, axis=0)

    tl.device_assert(mask_center != 0, 'mask center is zero!')
    tl.device_assert(tl.sum(skeleton_id_map == mask_center) > 0, 'not finding skeleton id')


    dummy = tl.zeros_like(skeleton_len_map)

    tl.device_assert(tl.sum(tl.where( skeleton_id_map==mask_center, skeleton_len_map, dummy)) > 0, 'where is busted')

    N_element_skel = tl.max(
        tl.where(skeleton_id_map == mask_center, skeleton_len_map, dummy)
    )

    tl.device_assert(N_element_skel > 0, 'skeleton n elements at id is zero')


    _off = tl.arange(0, SKEL_BLOCK_SIZE)

    skel_x_ptr = skeleton_pointer + _off * m_skeleton_stride + (0 * dim_skeleton_stride) + (index * n_skeleton_stride)
    skel_y_ptr = skeleton_pointer + _off * m_skeleton_stride + (1 * dim_skeleton_stride) + (index * n_skeleton_stride)
    skel_z_ptr = skeleton_pointer + _off * m_skeleton_stride + (2 * dim_skeleton_stride) + (index * n_skeleton_stride)

    # pid is the center px location.
    skl_x = tl.load(skel_x_ptr, _off < N_element_skel)
    skl_y = tl.load(skel_y_ptr, _off < N_element_skel)
    skl_z = tl.load(skel_z_ptr, _off < N_element_skel)

    dist = (
        (skl_x - x0) * (skl_x - x0) * dx
        + (skl_y - y0) * (skl_y - y0) * dy
        + (skl_z - z0) * (skl_z - z0) * dz
    )  # has shape of N_SKEL

    min_dist = tl.min(dist, axis=0)

    # out = tl.zeros((3, n_skeleton), dtype=tl.float16)
    _zeros = tl.zeros((SKEL_BLOCK_SIZE,), dtype=tl.float16)

    closest_x = tl.max(tl.where(dist == min_dist, skl_x, _zeros), axis=0)
    closest_y = tl.max(tl.where(dist == min_dist, skl_y, _zeros), axis=0)
    closest_z = tl.max(tl.where(dist == min_dist, skl_z, _zeros), axis=0)
    #
    # # store_x
    tl.store(
        out_pointer
        + (0 * out_c_stride)
        + (x0 * out_x_stride)
        + (y0 * out_y_stride)
        + (z0 * out_z_stride),
        closest_x,
    )

    # store_y
    tl.store(
        out_pointer
        + (1 * out_c_stride)
        + (x0 * out_x_stride)
        + (y0 * out_y_stride)
        + (z0 * out_z_stride),
        closest_y,
    )

    # store_z
    tl.store(
        out_pointer
        + (2 * out_c_stride)
        + (x0 * out_x_stride)
        + (y0 * out_y_stride)
        + (z0 * out_z_stride),
        closest_z,
    )


def next_power_of_2(x):
    return 1 if x == 0 else 2 ** (x - 1).bit_length()


def _bake_skeleton_triton(
    masks: Tensor,
    skeletons: Dict[int, Tensor],
    anisotropy: List[float],
    average: bool = True,
):
    """
    Launches a triton kernel to perform an in place distance calculation

    :param masks:
    :param skeletons: [N, 3]
    :param anisotropy:
    :param average:
    :return:
    """

    assert (
        masks.ndim == 3
    ), f"masks must have have 3 dimensions, not shape: {masks.shape}"
    assert masks.is_cuda, "masks must be on cuda"
    assert masks.is_contiguous(), "mask must be contiguous"
    for k, v in skeletons.items():
        assert v.is_cuda, "all skeletons must be on cuda device"
        assert v.is_contiguous, "all skeletons must be contiguous"
    assert len(anisotropy) == 3, "anisotropy should have 3 values"

    with torch.cuda.device(masks.device):
        x, y, z = masks.shape

        masks = masks.contiguous()
        baked = (
            torch.zeros((3, x, y, z)).to(masks.device).to(torch.float16).contiguous()
        )

        anisotropy = torch.tensor(anisotropy, device=masks.device).contiguous()

        skeleton_len_tensor = torch.zeros(len(skeletons), device=masks.device) # the size of each individual skeleton
        skeleton_id_map_tensor = torch.zeros(len(skeletons), device=masks.device) # the mapping of unique id to index in skeleton

        max_shape = max(v.shape[0] for v in skeletons.values()) if skeletons else 0
        if max_shape == 0:
            return baked

        combined_skeleton_tensors = torch.zeros((len(skeletons), max_shape, 3), device=masks.device) # [N_skel, M_points, 3]

        # Iterate over to place data in propper place...
        for i, (k,v) in enumerate(skeletons.items()):
            skeleton_len_tensor[i] = v.shape[0]  # place len of skel
            skeleton_id_map_tensor[i] =  k          # what index at each len
            combined_skeleton_tensors[i, 0:v.shape[0], :] = v

        assert skeleton_len_tensor.stride(0) == 1
        assert skeleton_id_map_tensor.stride(0) == 1


        num_out = 3
        grid = (x, y, z)

        N_skeletons = len(skeletons)
        dim_skel=3
        _min_skeleton_kernel[grid](
            mask_pointer=masks,
            skeleton_pointer=combined_skeleton_tensors,
            skeleton_id_map_pointer=skeleton_id_map_tensor,
            skeleton_len_pointer=skeleton_len_tensor,
            out_pointer=baked,
            anisotopy_pointer=anisotropy,
            mask_x_shape=x,
            mask_y_shape=y,
            mask_z_shape=z,
            mask_x_stride=masks.stride(0),
            mask_y_stride=masks.stride(1),
            mask_z_stride=masks.stride(2),
            out_c_stride=baked.stride(0),
            out_x_stride=baked.stride(1),
            out_y_stride=baked.stride(2),
            out_z_stride=baked.stride(3),
            out_c_shape=num_out,
            n_skeleton_stride=combined_skeleton_tensors.stride(0),
            m_skeleton_stride=combined_skeleton_tensors.stride(1),
            dim_skeleton_stride=combined_skeleton_tensors.stride(2),
            anisotropy_stride=anisotropy.stride(0),
            n_skeleton=N_skeletons,
            dim_skeleton=dim_skel,
            SKEL_BLOCK_SIZE=next_power_of_2(int(max_shape)),
            ID_BLOCK_SIZE=next_power_of_2(int(N_skeletons)),
        )
        torch.cuda.synchronize(masks.device)

    return baked


@torch.jit.ignore
def _bake_skeleton_torch(
    masks: Tensor,
    skeletons: Dict[int, Tensor],
    anisotropy: List[float] = (1.0, 1.0, 1.0),
    average: bool = True,
    device: str = "cpu",
) -> Tensor:
    r"""
    For each pixel :math:`p_ik` of object :math:`k` at index :math:`i\in[x,y,z]` in masks, returns a baked skeleton
    where the value at each index is the closest skeleton point :math:`s_{jk}` of any instance :math:`k`.

    This should reflect the ACTUAL spatial distance of your dataset for best results...These models tend to like XY
    embedding vectors more than Z. For anisotropic datasets, you should roughly provide the anisotropic correction
    factor of each voxel. For instance anisotropy of (1.0, 1.0, 5.0) means that the Z dimension is 5x larger than XY.


    Formally, the value at each position :math:`i\in[x,y,z]` of the baked skeleton tensor :math:`S` is the minimum of the
    euclidean distance function :math:`f(a, b)` and the skeleton point of any instance:

    .. math::
        S_{i} = min \left( f(i, s_{k})\right)\ for\ k \in [1, 2, ..., N]

    Shapes:
        - masks: :math:`(1, X_{in}, Y_{in}, Z_{in})`
        - skeletons: :math:`(3, N_i)`
        - anisotropy: :math:`(3)`
        - returns: :math:`(3, X_{in}, Y_{in}, Z_{in})`

    :param masks: Ground Truth instance mask of shape [1, X, Y, Z] of objects where each pixel is an integer id value.
    :param skeletons: Dict of skeleton indicies where each key is a unique instance of an object in mask.
        - Each skeleton has a shape [3, N] where N is the number of pixels constituting the skeleton
    :param anisotropy: Anisotropic correction factor for min distance calculation
    :param average: Average the skeletons such that there is a smooth transition form one area to the next
    :param device: torch.Device by which to run calculations

    :return: Baked skeleton
    """

    baked = torch.zeros(
        (3, masks.shape[0], masks.shape[1], masks.shape[2]), device=device
    )

    assert baked.device == masks.device, "Masks device must equal kwarg device"

    unique = torch.unique(masks)

    anisotropy = torch.tensor(anisotropy, device=device).view(1, 1, 3)

    for id in unique[unique != 0]:
        nonzero: Tensor = masks.eq(id).nonzero().unsqueeze(0).float()  # 1, N, 3
        skel: Tensor = skeletons[int(id)].to(device).unsqueeze(0).float()  # 1, N, 3

        # Calculate the distance between the skeleton of object 'id'
        # and all nonzero pixels of the binary mask of instance 'id'
        # print(skel.shape, nonzero.shape, id)

        # THIS IS SLOW!
        ind: Tensor = (
            torch.cdist(x1=skel.mul(anisotropy), x2=nonzero.mul(anisotropy))
            .squeeze(0)
            .argmin(dim=0)
        )

        nonzero = nonzero.squeeze(0).long()  # Can only index with long dtype

        baked[:, nonzero[:, 0], nonzero[:, 1], nonzero[:, 2]] = (
            skel[0, ind, :].float().T
        )
        del ind

    return baked


def bake_skeleton(
    masks: Tensor,
    skeletons: Dict[int, Tensor],
    anisotropy: List[float] = (1.0, 1.0, 1.0),
    average: bool = True,
    device: str = "cpu",
):
    r"""
    For each pixel :math:`p_ik` of object :math:`k` at index :math:`i\in[x,y,z]` in masks, returns a baked skeleton
    where the value at each index is the closest skeleton point :math:`s_{jk}` of any instance :math:`k`.

    This should reflect the ACTUAL spatial distance of your dataset for best results...These models tend to like XY
    embedding vectors more than Z. For anisotropic datasets, you should roughly provide the anisotropic correction
    factor of each voxel. For instance anisotropy of (1.0, 1.0, 5.0) means that the Z dimension is 5x larger than XY.


    Formally, the value at each position :math:`i\in[x,y,z]` of the baked skeleton tensor :math:`S` is the minimum of the
    euclidean distance function :math:`f(a, b)` and the skeleton point of any instance:

    .. math::
        S_{i} = min \left( f(i, s_{k})\right)\ for\ k \in [1, 2, ..., N]

    Shapes:
        - masks: :math:`(1, X_{in}, Y_{in}, Z_{in})`
        - skeletons: :math:`(3, N_i)`
        - anisotropy: :math:`(3)`
        - returns: :math:`(3, X_{in}, Y_{in}, Z_{in})`

    :param masks: Ground Truth instance mask of shape [1, X, Y, Z] of objects where each pixel is an integer id value.
    :param skeletons: Dict of skeleton indicies where each key is a unique instance of an object in mask.
        - Each skeleton has a shape [3, N] where N is the number of pixels constituting the skeleton
    :param anisotropy: Anisotropic correction factor for min distance calculation
    :param average: Average the skeletons such that there is a smooth transition form one area to the next
    :param device: torch.Device by which to run calculations

    :return: Baked skeleton
    """
    if -1 in skeletons:
        x, y, z = masks.shape
        baked = (
            torch.zeros((3, x, y, z)).to(masks.device).to(torch.float16).contiguous()
        )
        return baked


    if masks.is_cuda:
        logging.debug(f'skoots.lib.skeletons.bake_skeleton() | applying triton kernel for skeleton baking')
        if masks.ndim == 4 and masks.shape[0] == 1:
            masks = masks.squeeze(0).contiguous()
        baked = _bake_skeleton_triton(masks, skeletons, anisotropy)
    else:
        logging.debug(f'skoots.lib.skeletons.bake_skeleton() | applying pure torch kernel for skeleton baking')
        baked = _bake_skeleton_torch(masks, skeletons, anisotropy, device)

    logging.debug(f'skoots.lib.skeletons.bake_skeleton() | averaging baked skeletons')
    baked = average_baked_skeletons(baked.unsqueeze(0).float()).squeeze(0) if average else baked
    # baked[baked == 0] = -100

    return baked


def skeleton_to_mask(
    skeletons: Dict[int, Tensor],
    shape: Tuple[int, int, int],
    device: torch.device | str = None
) -> Tensor:
    r"""
    Converts a skeleton Dict to a skeleton mask which can simply be regressed against via Dice loss or similar...

    Shapes:
        - skeletons: [N, 3]
        - shape :math:`(3)`
        - returns: math:`(1, X_{in}, Y_{in}, Z_{in})`

    :param skeletons: Dict of skeletons
    :param shape: Shape of final mask
    :param device: output device
    :return: Mask of embedded skeleton px
    """

    skeleton_mask = None
    cached_inds = None

    if -1 in skeletons:
        return torch.zeros(shape, device)

    for k, v in skeletons.items():
        if skeleton_mask is None:
            skeleton_mask = torch.zeros(shape, device=v.device)
            cached_inds: Tensor = get_cached_disk_coords(device=v.device)

        # cached_inds = cached_inds[:, torch.logical_and(cached_inds[2,:].gt(-1), cached_inds[2,:].lt(1))]

        skeleton_indicies = (v.T.unsqueeze(1) + cached_inds.unsqueeze(2)).reshape(3, -1).long()
        # skeleton_indicies = v
        #
        x: Tensor = skeleton_indicies[0, :].long()
        y: Tensor = skeleton_indicies[1, :].long()
        z: Tensor = skeleton_indicies[2, :].long()

        ind_x = torch.logical_and(x >= 0, x < shape[0])
        ind_y = torch.logical_and(y >= 0, y < shape[1])
        ind_z = torch.logical_and(z >= 0, z < shape[2])

        ind = (
            ind_x.float() + ind_y.float() + ind_z.float()
        ) == 3  # Equivalent to a 3 way logical and

        skeleton_mask[
            skeleton_indicies[0, ind], skeleton_indicies[1, ind], skeleton_indicies[2, ind]
        ] = 1.0

    if skeleton_mask is None:  # in case skeletons has nothing in it
        skeleton_mask = torch.zeros(shape)

    return skeleton_mask.unsqueeze(0)


def _skeleton_to_mask(
    skeletons: Dict[int, Tensor],
    shape: Tuple[int, int, int],
    kernel_size: Tuple[int, int, int] = (15, 15, 1),
    n: int = 2,
) -> Tensor:
    r"""
    Converts a skeleton Dict to a skeleton mask which can simply be regressed against via Dice loss or whatever...

    Shapes:
        - skeletons: [N, 3]
        - shape :math:`(3)`
        - returns: math:`(1, X_{in}, Y_{in}, Z_{in})`

    :param skeletons: Dict of skeletons
    :param shape: Shape of final mask
    :return: Maks of embedded skeleton px
    """
    skeleton_mask = torch.zeros(shape)

    for k in skeletons:
        x = skeletons[k][:, 0].long()
        y = skeletons[k][:, 1].long()
        z = skeletons[k][:, 2].long()

        ind_x = torch.logical_and(x >= 0, x < shape[0])
        ind_y = torch.logical_and(y >= 0, y < shape[1])
        ind_z = torch.logical_and(z >= 0, z < shape[2])

        ind = (
            ind_x.float() + ind_y.float() + ind_z.float()
        ) == 3  # Equivalent to a 3 way logical and

        skeleton_mask[x[ind], y[ind], z[ind]] = 1

    skeleton_mask = skeleton_mask.unsqueeze(0).unsqueeze(0)

    for _ in range(
        n
    ):  # this might make things a bit better on the skeleton side of things...
        skeleton_mask = gauss_filter(
            binary_dilation_2d(skeleton_mask.gt(0.5).float()),
            kernel_size,
            [0.8, 0.8, 0.8],
        )

    return skeleton_mask.squeeze(0)


def index_skeleton_by_embed(skeleton: Tensor, embed: Tensor) -> Tensor:
    r"""
    Returns an instance mask by indexing skeleton with an embedding tensor
    For memory efficiency, skeleton is only ever Referenced! Never copied (I hope)

    Shapes:
        - skeleton: :math:`(B_{in}=1, 1, X_{in}, Y_{in}, Z_{in})`
        - embed: :math:`(B_{in}=1, 3, X_{in}, Y_{in}, Z_{in})`

    :param skeleton: Skeleton of a single instance
    :param embed: Embedding
    :return: torch.int instance mask
    """
    assert embed.device == skeleton.device, "embed and skeleton must be on same device"
    # assert embed.shape[2::] == skeleton.shape[2::], 'embed and skeleton must have identical spatial dimensions'
    assert (
        embed.ndim == 5 and skeleton.ndim == 5
    ), "Embed and skeleton must be a 5D tensor"

    b, c, x, y, z = embed.shape  # get the shape of the embedding
    embed = embed.view(
        (c, -1)
    ).round()  # flatten the embedding to extract it as an index

    # We need to only select indicies which lie within the skeleton
    x_ind = embed[0, :].clamp(0, skeleton.shape[2] - 1).long()
    y_ind = embed[1, :].clamp(0, skeleton.shape[3] - 1).long()
    z_ind = embed[2, :].clamp(0, skeleton.shape[4] - 1).long()

    out = torch.zeros(
        (1, 1, x, y, z), device=embed.device, dtype=torch.int
    ).flatten()  # For indexing to work, the out tensor
    # has to be flat

    ind = torch.arange(
        0, x_ind.shape[-1], device=embed.device
    )  # For each out pixel, we take the embedding at
    # that loc and assign it to skeleton

    out[ind] = skeleton[
        :, :, x_ind, y_ind, z_ind
    ].int()  # assign the skeleton ind to the out tensor

    return out.view(1, 1, x, y, z)  # return the re-shaped out tensor


if __name__ == "__main__":
    import skimage.io as io

    img = io.imread("/home/chris/Dropbox (Partners HealthCare)/skoots-experiments/data/plant-stem/train/0hrs_plant1_trim-acylYFP.tif")
    mask = io.imread("/home/chris/Dropbox (Partners HealthCare)/skoots-experiments/data/plant-stem/train/0hrs_plant1_trim-acylYFP.labels.tif")
    skel = torch.load("/home/chris/Dropbox (Partners HealthCare)/skoots-experiments/data/plant-stem/train/0hrs_plant1_trim-acylYFP.skeletons.trch")


    mask = torch.from_numpy(mask.astype(int)).contiguous().cuda()

    for k, v in skel.items():
        skel[k] = v.float().cuda()

    out = bake_skeleton(mask, skel, average=False)



    # n_mask = torch.zeros(len(skel)) # the size of each individual skeleton
    # ind_map = torch.zeros(len(skel)) # the mapping of unique id to index in skeleton
    #
    # max_shape = max(v.shape[0] for v in skel.values())
    #
    # all_skels = torch.zeros((len(skel), max_shape, 3))
    #
    # for i, (k,v) in enumerate(skel.items()):
    #     n_mask[i] = v.shape[0]
    #     ind_map[i] =  k
    #
    #     all_skels[i, 0:v.shape[0], :] = v
    #
    #
    # shape = img.shape
    # shape = (shape[1], shape[2], shape[0])
    # print(shape)
    #



import logging
import os.path
import tracemalloc
import warnings

import fastremap
import numpy as np
import skimage.io as io
import torch
import torch._dynamo
import torch.nn as nn
import zarr
from torch import Tensor
from torch.cuda.amp import autocast
from tqdm import tqdm
from yacs.config import CfgNode

import skoots.lib.skeleton
from skoots.lib.cropper import crops
from skoots.lib.flood_fill import efficient_flood_fill
from skoots.lib.morphology import binary_dilation, binary_dilation_2d
from skoots.lib.utils import cfg_to_bism_model
from skoots.lib.vector_to_embedding import vector_to_embedding

import time

warnings.filterwarnings("ignore")


@torch.inference_mode()  # disables autograd and reference counting for SPEED
def eval(
    image_path: str,
    checkpoint_path: str = "/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/models/Oct21_17-15-08_CHRISUBUNTU.trch",
) -> None:
    """
    Evaluates SKOOTS on an arbitrary image.

    :param image_path:
    :param checkpoint_path:
    :return:
    """
    tracemalloc.start()
    start = time.time()

    torch._dynamo.config.log_level = logging.ERROR
    logging.info(f"Loading model file: {checkpoint_path}")

    checkpoint = torch.load(checkpoint_path)
    if "cfg" in checkpoint:
        cfg: CfgNode = checkpoint["cfg"]
    else:
        raise RuntimeError("Attempting to evaluate skoots on a legacy model file.")

    device = "cuda:0" if torch.cuda.is_available() else "cpu"
    filename_without_extensions = os.path.splitext(image_path)[0]

    logging.info(f"Loading image from file: {image_path}")
    image: np.array = io.imread(image_path)  # [Z, X, Y, C]
    image: np.array = image[..., np.newaxis] if image.ndim == 3 else image
    image: np.array = image.transpose(-1, 1, 2, 0)
    image: np.array = image[[2], ...] if image.shape[0] > 3 else image  # [C=1, X, Y, Z]

    # if True:#image.dtype == torch.float:
    #     # pad3d = (5, 5, 30, 30, 30, 30)  # Pads last dim first!
    #     pad3d = ((0,0), (30, 30), (30, 30), (5, 5))
    #     image = np.pad(image, pad3d, mode='reflect') #if image.dtype == torch.float else image
    # else:
    #     pad3d = False

    c, x, y, z = image.shape
    logging.info(
        f"Loaded an image with shape: {(c, x, y, z)}, dtype: {image.dtype}, min: {image.min()}, max: {image.max()}"
    )

    scale: int = 2**16 if image.dtype == np.uint16 else 2**8

    image: Tensor = torch.from_numpy(image).to(torch.uint8).pin_memory()
    # print(f'Image Shape: {image.shape}, Dtype: {image.dtype}, Scale Factor: {scale}')

    # Allocate a bunch or things...

    vector_scale = torch.tensor(cfg.SKOOTS.VECTOR_SCALING)
    logging.info("Pre-allocating output arrays...")
    skeleton = torch.zeros(size=(1, x, y, z), dtype=torch.uint8)
    vectors = torch.zeros((3, x, y, z), dtype=torch.half)

    logging.info(f"Constructing SKOOTS model")
    base_model: nn.Module = cfg_to_bism_model(cfg)  # This is our skoots torch model
    base_model.load_state_dict(state_dict=checkpoint["model_state_dict"])
    base_model = base_model.to(device).train()

    logging.info(f"Compiling SKOOTS model with torch inductor")
    model = torch.compile(base_model)
    for _ in range(10):
        _ = model(torch.rand((1, 1, 300, 300, 20), device=device, dtype=torch.float))

    cropsize = [300, 300, 20]  # DEFAULT (300, 300, 20)
    overlap = [50, 50, 5]

    total = skoots.lib.cropper.get_total_num_crops(image.shape, cropsize, overlap)
    iterator = tqdm(
        crops(image, cropsize, overlap, device=device), desc="", total=total
    )
    benchmark_start = time.time()

    id = 1
    for crop, (x, y, z) in iterator:
        with autocast(enabled=True):  # Saves Memory!
            out = model(crop.div(scale).float().cuda())

        probability_map = out[:, [-1], ...]
        skeleton_map = out[:, [-2], ...].float()
        vec = out[:, 0:3:1, ...]

        vec = vec * probability_map.gt(0.5)
        skeleton_map = skeleton_map * probability_map.gt(0.5)

        for _ in range(
            2
        ):  # expand the skeletons in x/y/z. Only  because they can get too skinny
            skeleton_map = binary_dilation(skeleton_map)
            for _ in range(0):  # expand 2 times just in x/y
                skeleton_map = binary_dilation_2d(skeleton_map)

        # put the predictions into the preallocated tensors...
        _destination = (
            ...,
            slice(x + overlap[0], x + cropsize[0] - overlap[0]),
            slice(y + overlap[1], y + cropsize[1] - overlap[1]),
            slice(z + overlap[2], z + cropsize[2] - overlap[2]),
        )

        _source = (
            0,
            ...,
            slice(overlap[0], -overlap[0]),
            slice(overlap[1], -overlap[1]),
            slice(overlap[2], -overlap[2]),
        )

        skeleton[_destination] = skeleton_map[_source].gt(0.8).cpu()
        vectors[_destination] = vec[_source].half().cpu()

        iterator.desc = f"Evaluating UNet on slice [x{x}:y{y}:z{z}]"

    # _x, _y, _z
    # if pad3d:
    #     _, padx, pady, padz = pad3d
    #
    #     skeleton = skeleton[0, padx[0]:-padx[1], pady[0]:-pady[1], padz[0]:-padz[1]]
    #     vectors = vectors[:, padx[0]:-padx[1], pady[0]:-pady[1], padz[0]:-padz[1]]
    #
    # else:
    #     skeleton = skeleton[0, ...]
    #     vectors = vectors[:, ...]

    del image  # we don't need the image anymore

    # torch.save(vectors, filename_without_extensions + '_vectors.trch')
    # torch.save(skeleton, filename_without_extensions + '_unlabeled_skeletons.trch')

    # logging.info(f"Saving vector output")
    # io.imsave(
    #     filename_without_extensions + f"_skoots_vectors.tif",
    #     vectors.mul(127)
    #     .add(127)
    #     .int()
    #     .cpu()
    #     .numpy()
    #     .transpose(3, 1, 2, 0)
    #     .astype(np.uint16),
    #     compression="zlib",
    # )
    #
    # logging.info(f"Saving unlabeledc skeleton output")
    # io.imsave(
    #     filename_without_extensions + f"_unlabeled_skeletons.tif",
    #     skeleton.mul(255)
    #     .squeeze(0)
    #     .clamp(0, 255)
    #     .int()
    #     .cpu()
    #     .numpy()
    #     .transpose(2, 0, 1)
    #     .astype(np.uint8),
    #     compression="zlib",
    # )

    zarr.save_array(
        filename_without_extensions + f"_skoots_vectors.zarr",
        vectors.mul(127).add(127).int().cpu().numpy(),
    )

    zarr.save_array(
        filename_without_extensions + f"_skoots_skeleton_unlabeled.zarr",
        skeleton.cpu().numpy(),
    )

    logging.info(f"Performing an flood fill on skeletons")
    skeleton: Tensor = efficient_flood_fill(skeleton.to(torch.int16))

    # io.imsave(
    #     filename_without_extensions + f"_skeletons.tif",
    #     skeleton
    #     .squeeze(0)
    #     .clamp(0, 255)
    #     .int()
    #     .cpu()
    #     .numpy()
    #     .transpose(2, 0, 1)
    #     .astype(np.uint16),
    #     compression="zlib",
    # )

    zarr.save_array(
        filename_without_extensions + f"_skoots_skeleton.zarr", skeleton.cpu().numpy()
    )

    logging.info(f"Saving labeled skeletons")
    # torch.save(skeleton, filename_without_extensions + '_skeletons.trch')

    instance_mask = torch.zeros_like(skeleton, dtype=torch.int16)
    skeleton = skeleton.unsqueeze(0).unsqueeze(0)

    cropsize = [500, 500, 50]
    overlap = (50, 50, 5)
    total = skoots.lib.cropper.get_total_num_crops(vectors.shape, cropsize, overlap)
    iterator = tqdm(
        crops(vectors, crop_size=cropsize, overlap=overlap),
        desc="Assigning Instances:",
        total=total,
    )

    logging.info(f"Identifying connected components...")
    for _vec, (x, y, z) in iterator:
        _destination = (
            slice(x + overlap[0], x + cropsize[0] - overlap[0]),
            slice(y + overlap[1], y + cropsize[1] - overlap[1]),
            slice(z + overlap[2], z + cropsize[2] - overlap[2]),
        )

        _source = (
            slice(overlap[0], -overlap[0]),
            slice(overlap[1], -overlap[1]),
            slice(overlap[2], -overlap[2]),
        )

        _embed = skoots.lib.vector_to_embedding.vector_to_embedding(
            scale=vector_scale, vector=_vec, N=1
        )
        _embed += torch.tensor((x, y, z)).view(
            1, 3, 1, 1, 1
        )  # We adjust embedding to region of the crop
        _inst_maks = skoots.lib.skeleton.index_skeleton_by_embed(
            skeleton=skeleton, embed=_embed
        ).squeeze()
        w, h, d = _inst_maks.shape

        instance_mask[_destination] = (
            _inst_maks[_source] if torch.tensor(overlap).gt(0).all() else _inst_maks
        )

    with open(filename_without_extensions + "_skoots_benchmark.txt", "w") as f:
        ss = tracemalloc.get_traced_memory()
        f.write(f"SKOOTS Segmentation Benchmark:\n")
        f.write(f"------------------------------\n")
        f.write(f"Time: {time.time() - benchmark_start} seconds\n")
        f.write(f"Memory (current/max): {tracemalloc.get_traced_memory()}\n\n")

    print("DONE")
    # print(instance_mask.unique().shape[0] - 1, ' Unique mito')

    del skeleton, vectors  # explicitly delete unnecessary tensors for memory
    instance_mask, remapping = fastremap.renumber(
        instance_mask.cpu().numpy(), in_place=True
    )

    io.imsave(filename_without_extensions + '_instance_mask.tif',
              instance_mask.transpose(2, 0, 1), compression='zlib')
    # zarr.save_array(
    #     filename_without_extensions + f"_skoots_instance_mask.zarr", instance_mask
    # )

    end = time.time()
    elapsed = end - start

    logging.info(
        f"DONE: Process took {elapsed} seconds, {elapsed / 60} minutes, {elapsed / (60 ** 2)}, hours"
    )


if __name__ == "__main__":
    # image_path = '/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/outputs/hide_validate-1.tif'
    image_path = "/home/chris/Documents/threeOHC_registered_8bit_cell2.tif"
    # image_path = '/home/chris/Dropbox (Partners HealthCare)/Manuscripts - Buswinka/Mitochondria Segmentation/Figures/Fig X - compare to affinity/data/hide_validate.tif'
    # image_path = '/home/chris/Documents/threeOHC_registered_8bit.tif'
    # image_path = '/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/outputs/onemito.tif'
    # image_path = '/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/outputs/cell_apex-1.tif'
    # image_path = '/home/chris/Dropbox (Partners HealthCare)/Manuscripts - Buswinka/Mitochondria Segmentation/Figures/Figure X6X  - Whole image analysis/crop.tif'
    import time

    t1 = time.time()
    eval(image_path)
    t2 = time.time()

    print(f"TOOK {t1 - t2} seconds")



import glob
import os.path

import numpy as np
import skimage.io as io
import torch
import zarr
from torch import Tensor


def convert(base_dir):
    if os.path.isdir(base_dir) and not base_dir.endswith(".zarr"):
        files = glob.glob(os.path.join(base_dir, "*.trch"))

    elif '*' in base_dir:
        files = glob.glob(base_dir)
    else:
        files = [base_dir]
    print("Found the following files to convert:")
    for f in files:
        if os.path.exists(f):
            print(f"\t{f}")
        # print(f'\t{f}')
    print("------")
    for f in files:
        if not os.path.exists(f):
            continue
        print(f"Converting {f}...", end="")

        if f.endswith(".zarr"):
            x = zarr.load(f)
        else:
            x = torch.load(f, map_location="cpu")

        print(x.shape)

        if not isinstance(x, Tensor) and not isinstance(x, np.ndarray):
            continue

        new_file = os.path.splitext(f)[0] + ".tif"
        if isinstance(x, np.ndarray):
            if x.ndim == 3:
                x = x.transpose(2, 0, 1)
                io.imsave(new_file, x, compression="zlib")

            elif x.ndim == 4:
                x = x.transpose(3, 1, 2, 0).astype(np.uint16)
                io.imsave(new_file, x, compression="zlib")

        else:
            if x.min() < 0:
                print("doing this...", x.min())
                mask = x == 0
                x = (
                    x.add(1).div(2).mul(255).float().round().to(torch.uint8)
                )  # convert skeletons to uint8
                x[mask] = 0
                print(x.shape, x.min(), x.max())

            if x.ndim == 3:
                io.imsave(new_file, x.permute(2, 0, 1).numpy(), compression="zlib")
            elif x.ndim == 4:
                print(new_file)
                io.imsave(
                    new_file, x.permute(3, 1, 2, 0).numpy(), compression="zlib"
                )  # Z, C, X, Y

        del x



from typing import Optional

import numpy as np
import skimage.io as io
import torch
from torch import Tensor


def imread(image_path: str, pin_memory: Optional[bool] = False) -> Tensor:
    """
    Imports an image from file and returns in torch format

    :param image_path: path to image
    :param pin_memory: saves torch tensor in pinned memory if true
    :return:
    """

    image: np.array = io.imread(image_path)  # [Z, X, Y, C]
    image: np.array = image[..., np.newaxis] if image.ndim == 3 else image
    image: np.array = image.transpose(-1, 1, 2, 0)
    image: np.array = image[[2], ...] if image.shape[0] > 3 else image  # [C=1, X, Y, Z]

    image: Tensor = torch.from_numpy(image.astype(np.int32))

    if pin_memory:
        image: Tensor = image.pin_memory()

    return image


if __name__ == "__main__":
    from skoots.config import get_cfg_defaults

    cfg = get_cfg_defaults()
    cfg_to_bism_model(cfg)



import argparse
import glob
import logging
import os.path

import skoots.lib.eval
import skoots.train.generate_skeletons
import skoots.utils.convert_trch_to_tif


def main():
    parser = argparse.ArgumentParser(
        prog="SKOOTS",
        description="skoots parameters",
    )

    # general script arguments
    eval_args = parser.add_argument_group("eval arguments")
    eval_args.add_argument("--image", type=str, help="path to image")
    eval_args.add_argument(
        "--pretrained-checkpoint",
        type=str,
        help="path to a pretrained skoots model. Will be used"
        "as a starting point for training",
    )

    eval_args.add_argument(
        "--log",
        type=int,
        default=3,
        help="Log Level: 0-Debug, 1-Info, 2-Warning, 3-Error, 4-Critical",
    )

    _log_map = [
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]

    # accessory script arguments
    accessory_args = parser.add_argument_group("scripting arguments")
    accessory_args.add_argument(
        "--skeletonize-train-data", help="calculate skeletons of training data"
    )

    accessory_args.add_argument(
        "--mask-filter", default='.labels', help="filter of mask file"
    )

    accessory_args.add_argument(
        "--anisotropyXY",
        type=float,
        default=1.0,
        help="calculate skeletons of training data",
    )
    accessory_args.add_argument(
        "--anisotropyZ",
        type=float,
        default=1.0,
        help="calculate skeletons of training data",
    )
    accessory_args.add_argument(
        "--convert",
        type=str,
        help="converts all skoots eval outputs in directory to a tif image",
    )

    args = parser.parse_args()

    logging.basicConfig(
        level=_log_map[args.log],
        format="[%(asctime)s] skoots-eval [%(levelname)s]: %(message)s",
    )

    # Eval
    if args.skeletonize_train_data is None and args.convert is None:
        assert args.pretrained_checkpoint is not None, (
            f"Cannot evaluate SKOOTS wihtout pretrained model. "
            f"--pretrained_checkpoint must not be None"
        )
        if args.pretrained_checkpoint == 'base':
            args.pretrained_checkpoint = "/home/chris/Dropbox (Partners HealthCare)/skoots-experiments/models/mito/Mar28_16-26-14_CHRISUBUNTU.trch"

        if os.path.isdir(args.image):
            files = glob.glob(args.image + "/*.tif")
            files.sort()
        else:
            files = [args.image]
        for f in files:
            skoots.lib.eval.eval(f, args.pretrained_checkpoint)

    # accessory scripts
    if args.skeletonize_train_data:
        downscale = (args.anisotropyXY, args.anisotropyXY, args.anisotropyZ)
        skoots.train.generate_skeletons.create_gt_skeletons(
            args.skeletonize_train_data, args.mask_filter, downscale
        )

    if args.convert:
        skoots.utils.convert_trch_to_tif.convert(args.convert)


if __name__ == "__main__":
    main()



from typing import Optional, List, Union

import skimage.io
import skimage.measure
import torch
import torch.nn
from fvcore.nn import FlopCountAnalysis
from fvcore.nn.parameter_count import parameter_count
from torch import Tensor


def get_volume(
    x: Tensor, spacing: Optional[Union[List[float], Tensor]] = None
) -> Tensor:
    """
    Returns the volume of all nonzero voxels in the tensor

    :param x: [X, Y, Z] bool tensor of an image
    :param spacing: Array of the voxel spacings with which to calculate the volume

    :return: Volume
    """

    volume = torch.sum(x.nonzero().shape[0])
    volume = volume * torch.tensor(spacing).prod() if spacing else volume

    return volume


def get_surface_area(x: Tensor, anisotropy_ratio: List[float]) -> Tensor:
    """
    Approximates object surface area by performing a marching cubes algorithm to get a collection of faces.
    Then calculates surface area of all polygons.

    :param x: [X, Y, Z] bool tensor of an image
    :param anisotropy_ratio:

    :return: surface area in arbitrary units.
    """

    vets, faces, normals, values = skimage.measure.marching_cubes(
        x.gt(0).mul(255).cpu().numpy(), spacing=anisotropy_ratio
    )

    surface_area = skimage.measure.mesh_surface_area(verts=vets, faces=faces)
    surface_area = torch.from_numpy(surface_area)

    return surface_area


def tversky(x: Tensor, y: Tensor, a: Tensor, b: Tensor) -> Tensor:
    """

    :param x:
    :param y:
    :param a: Penalizes false positives
    :param b: Penalizes false negatives
    :return:
    """

    true_positive: Tensor = x.mul(y).sum()
    false_positive: Tensor = torch.logical_not(x).mul(y).sum().mul(a)
    false_negative: Tensor = ((1 - y) * x).sum() * b

    tversky = (true_positive + 1e-16) / (
        true_positive + false_positive + false_negative + 1e-16
    )

    return 1 - tversky


def get_flops(model: torch.nn.Module, example_input: Tensor):
    return FlopCountAnalysis(model, example_input).total()


def get_parameter_count(model: torch.nn.Module):
    param_dict = parameter_count(model)

    total = 0
    for k, v in param_dict.items():
        total += v

    return total


if __name__ == "__main__":
    from hcat.backends.unext_hydra import UNeXT as UNeXT_0
    from hcat.backends.unext import UNeXT as UNeXT_1
    import matplotlib.pyplot as plt

    f = []
    p = []
    for i in (5, 7, 9):
        model0 = UNeXT_0(depths=[2, 3, 4, 3, 2], kernel_size=i)
        model0 = model0.to("cuda:0")

        flops0 = get_flops(model0, torch.rand((1, 1, 256, 256, 10), device="cuda:0"))
        total_params0 = get_parameter_count(model0)
        f.append(flops0)
        p.append(total_params0)

    plt.plot((5, 7, 9), f)

    f = []
    p = []
    for i in (5, 7, 9):
        model1 = UNeXT_1(depths=[2, 3, 4, 3, 2], kernel_size=i)
        model1 = model1.to("cuda:0")

        flops1 = get_flops(model1, torch.rand((1, 1, 256, 256, 10), device="cuda:0"))
        total_params1 = get_parameter_count(model1)
        f.append(flops1)
        p.append(total_params1)

    plt.plot((5, 7, 9), f)
    plt.show()



from typing import List, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from skoots.lib.morphology import binary_erosion
from skoots.lib.utils import crop_to_identical_size


def tversky_graphable(pred, gt, alpha, beta):
    true_positive: Tensor = pred.mul(gt).sum()
    false_positive: Tensor = torch.logical_not(gt).mul(pred).sum().add(1e-10).mul(alpha)
    false_negative: Tensor = ((1 - pred) * gt).sum() * beta

    tversky = (true_positive + 1e-10) / (
        true_positive + false_positive + false_negative + 1e-10
    )

    return 1 - tversky


class jaccard(nn.Module):
    def __init__(self):
        super(jaccard, self).__init__()

    def forward(
        self, predicted: torch.Tensor, ground_truth: torch.Tensor, eps: float = 1e-10
    ) -> torch.Tensor:
        """
        Returns jaccard index of two torch.Tensors

        :param predicted: [B, I, X, Y, Z] torch.Tensor
                - probabilities calculated from hcat.utils.embedding_to_probability
                  where B: is batch size, I: instances in image
        :param ground_truth: [B, I, X, Y, Z] torch.Tensor
                - segmentation mask for each instance (I).
        :param eps: float
                - Very small number to ensure numerical stability. Default 1e-10
        :return: jaccard_loss: [1] torch.Tensor
                - Result of Loss Function Calculation
        """

        # Crop both tensors to the same shape
        predicted, ground_truth = crop_to_identical_size(predicted, ground_truth)

        intersection = (predicted * ground_truth).sum().add(eps)
        union = (predicted + ground_truth).sum().sub(intersection).add(eps)

        return 1.0 - (intersection / union)


class dice(nn.Module):
    def __init__(self):
        super(dice, self).__init__()

    def forward(
        self, predicted: torch.Tensor, ground_truth: torch.Tensor, smooth: float = 1e-10
    ) -> torch.Tensor:
        """
        Returns dice index of two torch.Tensors

        :param predicted: [B, I, X, Y, Z] torch.Tensor
                - probabilities calculated from hcat.utils.embedding_to_probability
                  where B: is batch size, I: instances in image
        :param ground_truth: [B, I, X, Y, Z] torch.Tensor
                - segmentation mask for each instance (I).
        :param smooth: float
                - Very small number to ensure numerical stability. Default 1e-10
        :return: dice_loss: [1] torch.Tensor
                - Result of Loss Function Calculation
        """

        # Crop both tensors to the same shape
        predicted, ground_truth = crop_to_identical_size(predicted, ground_truth)

        intersection = (predicted * ground_truth).sum().add(smooth)
        denominator = (predicted + ground_truth).sum().add(smooth)
        loss = 2 * intersection / denominator

        return 1 - loss


class tversky(nn.Module):
    def __init__(self, alpha: float, beta: float, eps: float):
        """
        Returns dice index of two torch.Tensors

        :param alpha: float
                - Value which penalizes False Positive Values
        :param beta: float
                - Value which penalizes False Negatives
        :param eps: float
                - Numerical stability term
        """
        super(tversky, self).__init__()

        self.alpha = torch.tensor(float(alpha))
        self.beta = torch.tensor(float(beta))
        self.eps = torch.tensor(float(eps))

    def forward(
        self, predicted: Union[Tensor, List[Tensor]], ground_truth: Tensor
    ) -> Tensor:
        if self.alpha.device != predicted.device:  # silently caches device
            self.alpha.to(predicted.device)
            self.beta.to(predicted.device)
            self.eps.to(predicted.device)

        futures: List[torch.jit.Future[torch.Tensor]] = []

        # List of Tensors
        if isinstance(predicted, list):
            for i, pred in enumerate(predicted):
                futures.append(
                    torch.jit.fork(
                        self._tversky,
                        pred,
                        ground_truth[i, ...],
                        self.alpha,
                        self.beta,
                        self.eps,
                    )
                )

        # Already Batched Tensor
        elif isinstance(predicted, Tensor):
            for i in range(predicted.shape[0]):
                futures.append(
                    torch.jit.fork(
                        self._tversky,
                        predicted[i, ...],
                        ground_truth[i, ...],
                        self.alpha,
                        self.beta,
                        self.eps,
                    )
                )

        results: List[Tensor] = []
        for future in futures:
            results.append(torch.jit.wait(future))

        return torch.mean(torch.stack(results))

    @staticmethod
    def _tversky(
        pred: Tensor, gt: Tensor, alpha: Tensor, beta: Tensor, eps: float = 1e-8
    ):
        """
        tversky loss on per image basis.


        Args:
            pred: [N, X, Y, Z] Tensor of predicted segmentation masks (N instances)
            gt: [N, X, Y, Z] Tensor of ground truth segmentation masks (N instances)
            alpha: Penalty to false positives
            beta: Penalty to false negatives
            eps: stability parameter

        Returns:

        """

        # ------------------- Expand Masks
        unique = torch.unique(gt)
        unique = unique[unique != 0]

        # assert gt.ndim == 4, f'{gt.shape=}'

        _, x, y, z = gt.shape
        nd_masks = torch.zeros((unique.shape[0], x, y, z), device=pred.device)
        for i, id in enumerate(unique):
            nd_masks[i, ...] = (gt == id).float().squeeze(0)

        pred, nd_masks = crop_to_identical_size(pred, nd_masks)

        # assert not torch.any(torch.isnan(pred)), torch.sum(torch.isnan(pred))

        true_positive: Tensor = pred.mul(nd_masks).sum()
        false_positive: Tensor = (
            torch.logical_not(nd_masks).mul(pred).sum().add(1e-10).mul(alpha)
        )
        false_negative: Tensor = ((1 - pred) * nd_masks).sum() * beta

        # assert not torch.any(torch.isnan(true_positive)), torch.sum(torch.isnan(true_positive))
        # assert not torch.any(torch.isnan(false_negative))
        # assert not torch.any(torch.isnan(false_positive))
        #
        # assert not torch.any(torch.isinf(true_positive))
        # assert not torch.any(torch.isinf(false_negative))
        # assert not torch.any(torch.isinf(false_positive))

        tversky = (true_positive + eps) / (
            true_positive + false_positive + false_negative + eps
        )

        return 1 - tversky

    def __repr__(self):
        return f"LossFn[name=tversky, alpha={self.alpha.item()}, beta={self.beta.item()}, eps={self.eps.item()}"


class split(nn.Module):
    def __init__(self, n_iter: int = 2, alpha: float = 2.0, device: str = "cpu"):
        """
        The "oh shit my skeletons have split" loss. This basically checks if an edge has crossed the middle
        of a GT object. If it does, it applies a crazy loss.

        Approximates the distance function by just eroding and adding a bunch.
        Approximates the edge function by subtracting the prediction by the eroded

        For speed, will only check for pixels :math:`n_{iter}` away. So if :math:`n_{iter} = 3' the maximum distance
        any pixel might be from an edge would be 3.

        Formally:
            if :math:`E` is the edge function and :math:`\Phi` is the distance function, we compute the loss :math:`L(s, p)`
            where :math:`s` is the ground truth skeleton, and :math:`p` is the predicted skeleton where

        .. math::
            L(s, p) = E(p)^{ \alpha \Phi(s)} - 1


        :param n_iter: Number of times to perform erosion for distance calculation.
        :param alpha: Scale factor for exponential loss. Large values penalize breakages more.
        :param device: a torch.device - 'cuda' or 'cpu'
        """

        super(split, self).__init__()
        self.n = n_iter
        self._device = device
        self.a = torch.tensor(alpha, device=device)

    def forward(self, pred, gt):
        distance = torch.zeros_like(gt) + gt

        for _ in range(self.n):
            gt = binary_erosion(gt)
            distance = distance + gt  # psuedo distance function...

        distance = distance.div(self.n - 1)

        pred = pred.sub(binary_erosion(pred)).mul(2)  # cheeky edge detection function

        _split_loss = self._split_loss(edges=pred, distance=distance, a=self.a)

        return _split_loss.mean()

    @staticmethod
    # @torch.jit.script
    def _split_loss(edges: Tensor, distance: Tensor, a: Tensor):
        return torch.pow(edges, a * distance)


############## clDICE LOSS FROM: https://github.com/jocpae/clDice/blob/master/cldice_loss/pytorch/soft_skeleton.py


def soft_erode(img: Tensor) -> Tensor:
    """approximates morphological operations through max_pooling for 2D and 3D"""
    if len(img.shape) == 4:
        p1 = -F.max_pool2d(-img, (3, 1), (1, 1), (1, 0))
        p2 = -F.max_pool2d(-img, (1, 3), (1, 1), (0, 1))
        return torch.min(p1, p2)
    elif len(img.shape) == 5:
        p1 = -F.max_pool3d(-img, (3, 1, 1), (1, 1, 1), (1, 0, 0))
        p2 = -F.max_pool3d(-img, (1, 3, 1), (1, 1, 1), (0, 1, 0))
        p3 = -F.max_pool3d(-img, (1, 1, 3), (1, 1, 1), (0, 0, 1))
        return torch.min(torch.min(p1, p2), p3)


def soft_dilate(img: Tensor) -> Tensor:
    """approximates morphological operations through max_pooling for 2D and 3D"""
    if len(img.shape) == 4:
        return F.max_pool2d(img, (3, 3), (1, 1), (1, 1))
    elif len(img.shape) == 5:
        return F.max_pool3d(img, (3, 3, 3), (1, 1, 1), (1, 1, 1))


def soft_open(img: Tensor) -> Tensor:
    """approximates morphological operations through max_pooling for 2D and 3D"""
    return soft_dilate(soft_erode(img))


def soft_skeletonize(img: Tensor, iter_: int) -> Tensor:
    """
    Performs a soft-skeletonization by terativly performing "soft morphological operations"

    :param img: Image to perform operation on
    :param iter_: Number of times to perform the operation
    :return: Soft-skeleton
    """
    img1 = soft_open(img)
    skel = F.relu(img - img1)
    for j in range(iter_):
        img = soft_erode(img)
        img1 = soft_open(img)
        delta = F.relu(img - img1)
        skel = skel + F.relu(delta - skel * delta)
    return skel


class soft_cldice(nn.Module):
    def __init__(self, iter_=3, smooth=1.0):
        super(soft_cldice, self).__init__()
        self.iter = iter_
        self.smooth = smooth

    def forward(self, predicted: Tensor, ground_truth: Tensor) -> Tensor:
        """
        Calculates the soft-clDice metric on a true and predicted value

        :param ground_truth:
        :param predicted:
        :return:
        """
        skeleton_predicted = soft_skeletonize(predicted, self.iter)
        skeleton_true = soft_skeletonize(ground_truth, self.iter)

        tprec = (
            torch.sum(torch.multiply(skeleton_predicted, ground_truth)[:, 1:, ...])
            + self.smooth
        ) / (torch.sum(skeleton_predicted[:, 1:, ...]) + self.smooth)
        tsens = (
            torch.sum(torch.multiply(skeleton_true, predicted)[:, 1:, ...])
            + self.smooth
        ) / (torch.sum(skeleton_true[:, 1:, ...]) + self.smooth)

        cl_dice = 1.0 - 2.0 * (tprec * tsens) / (tprec + tsens)

        return cl_dice


def soft_dice(predicted: Tensor, ground_truth: Tensor, smooth: int = 1) -> Tensor:
    """
    Computes the soft dice metric

    :param ground_truth:
    :param predicted:
    :param smooth: smoothing factor to prevent division by zero
    :return:
    """
    intersection = torch.sum((ground_truth * predicted))
    coeff = (2.0 * intersection + smooth) / (
        torch.sum(ground_truth) + torch.sum(predicted) + smooth
    )

    return 1.0 - coeff


class soft_dice_cldice(nn.Module):
    def __init__(self, iter_=3, alpha=0.5, smooth=1.0):
        super(soft_dice_cldice, self).__init__()
        self.iter = iter_
        self.smooth = smooth
        self.alpha = alpha

    def forward(self, predicted: Tensor, ground_truth: Tensor) -> Tensor:
        """
        Calculates a singular loss value combining soft-Dice and soft-clDice which can be used to train
        a neural network

        :param predicted: Input tensor
        :param ground_truth: Ground Truth Tensor
        :return: Single value which to perform a backwards pass
        """

        dice = soft_dice(ground_truth, predicted)

        skel_pred = soft_skeletonize(predicted, self.iter)
        skel_true = soft_skeletonize(ground_truth, self.iter)

        tprec = (torch.sum(skel_pred * ground_truth) + self.smooth) / (
            torch.sum(skel_pred) + self.smooth
        )
        tsens = (torch.sum(skel_true * predicted) + self.smooth) / (
            torch.sum(skel_true) + self.smooth
        )

        cl_dice = 1.0 - 2.0 * (tprec * tsens) / (tprec + tsens)

        return (1.0 - self.alpha) * dice + self.alpha * cl_dice


if __name__ == "__main__":
    lossfn = soft_dice_cldice()

    predicted = torch.rand((1, 1, 20, 20, 10), device="cpu")
    gt = torch.rand((1, 1, 20, 20, 10), device="cpu").round().float()
    a = lossfn(predicted, gt)
    print(a)



import argparse
import logging
import os.path

import matplotlib.pyplot as plt
import torch

from skoots.validate.lib import (
    get_segmentation_errors,
    mask_iou,
    accuracies_from_iou,
    f1_score,
    mask_dice,
    mask_soft_cldice,
)
from skoots.validate.utils import imread


def main() -> None:
    """
    Runs validation stats on two isntance masks. GT was manually calculated, pred is predicted by an algorithm

    :param gt:
    :param pred:
    :return:
    """
    parser = argparse.ArgumentParser(description="SKOOTS Training Parameters")
    parser.add_argument(
        "--ground_truth", type=str, help="Path to ground truth instance mask"
    )
    parser.add_argument(
        "--predicted", type=str, help="Path to ground truth instance mask"
    )
    parser.add_argument(
        "--log",
        type=int,
        default=3,
        help="Log Level: 0-Debug, 1-Info, 2-Warning, 3-Error, 4-Critical",
    )
    args = parser.parse_args()

    _log_map = [
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]
    logging.basicConfig(
        level=_log_map[args.log],
        format="[%(asctime)s] bism-eval [%(levelname)s]: %(message)s",
    )

    gt_path, pred_path = args.ground_truth, args.predicted

    # Load model
    if os.path.exists(gt_path) and os.path.exists(pred_path):
        file_without_extension = os.path.splitext(pred_path)[0]
        gt, pred = imread(gt_path), imread(pred_path)

        gt = gt[:, 50:-50, 50:-50, 5:-5]
        pred = pred[:, 50:-50, 50:-50, 5:-5]

        logging.debug(f"Ground Truth Shape: {gt.shape}, Predicted Shape: {pred.shape}")
    else:
        raise RuntimeError(f"{os.path.exists(gt_path)=}, {os.path.exists(pred_path)=}")

    # sent to device for speed
    device = "cuda" if torch.cuda.is_available() else "cpu"

    gt = gt.to(device)
    pred = pred.to(device)

    print("Calculating Segmentation Errors...")
    over_segmentation_rate, under_segmentation_rate = get_segmentation_errors(gt, pred)

    print("Calculating Instance Intersection over Union...")
    iou = mask_iou(gt, pred)
    dice = mask_dice(gt, pred)
    cldice = mask_soft_cldice(gt, pred)

    print("Calculating Accuracy Statistics...")
    tfp = [accuracies_from_iou(iou, thr / 100) for thr in range(100)]
    precision = [(tp / (tp + fp)) for (tp, fp, fn) in tfp]
    recall = [(tp / (tp + fn)) for (tp, fp, fn) in tfp]
    f1 = [f1_score(*a) for a in tfp]

    # Precision
    _x = torch.arange(0, 100).numpy()
    fig = plt.figure()
    plt.plot(_x, precision, "k-")
    plt.title("Precision")
    plt.xlabel("Threshold (%)")
    plt.ylabel("Score")
    plt.tight_layout()
    plt.savefig(f"{file_without_extension}_precision.png", dpi=300)
    plt.close()

    # Recall
    _x = torch.arange(0, 100).numpy()
    fig = plt.figure()
    plt.plot(_x, recall, "k-")
    plt.title("Precision")
    plt.xlabel("Threshold (%)")
    plt.ylabel("Score")
    plt.tight_layout()
    plt.savefig(f"{file_without_extension}_recall.png", dpi=300)
    plt.close()

    # F1
    _x = torch.arange(0, 100).numpy()
    fig = plt.figure()
    plt.plot(_x, f1, "k-")
    plt.title("F1 Score")
    plt.xlabel("Threshold (%)")
    plt.ylabel("Score")
    plt.tight_layout()
    plt.savefig(f"{file_without_extension}_f1.png", dpi=300)
    plt.close()

    print("Writing File...")
    with open(f"{file_without_extension}_accuracy_stats.csv", "w") as file:
        file.write(f"Ground Truth File: {gt_path}\n")
        file.write(f"Predicted File: {pred_path}\n")
        file.write(f"Over Segmentation Rate: {over_segmentation_rate}\n")
        file.write(f"Under Segmentation Rate: {under_segmentation_rate}\n")
        file.write(
            f"thr,true_positive,false_positive,false_negative,precision,recall,f1\n"
        )
        i = -1
        for (tp, fp, fn), _precision, _recall, _f1 in zip(tfp, precision, recall, f1):
            i += 1
            file.write(f"{i / 100},{tp},{fp},{fn},{_precision},{_recall},{_f1}\n")

    print(f"File Written: {file_without_extension}_accuracy_stats.csv")

    with open(f"{file_without_extension}_intersection_over_union.csv", "w") as file:
        file.write(f"Ground Truth File: {gt_path}\n")
        file.write(f"Predicted File: {pred_path}\n")
        file.write(f"Average IOU: {iou.max(1)[0].mean().item()}\n")
        file.write(f"Average Dice: {dice.max(1)[0].mean().item()}\n")
        file.write(f"Average clDice: {cldice.max(1)[0].mean().item()}\n")
        file.write(f"gt_label,best_iou,best_dice,best_cldice\n")

        for i, u in enumerate(gt.unique()):
            if u == 0:
                continue
            file.write(
                f"{u},{iou[i - 1, :].max().item()},{dice[i - 1, :].max().item()},{cldice[i - 1, :].max().item()}\n"
            )

    print(f"File Written: {file_without_extension}_intersection_over_union.csv")



from typing import List
from typing import Tuple

import torch
from torch import Tensor

"""
2D and 3D implementations of vector to embedding. 

vector is predicted by a network and is an array of values ranging from -1 to 1. These values can be scaled by
"scale" and are used to project pixels from their location, to a new location in embedding space. 

Functionally, a trained network will push pixels to the center of an object, forming clusers which can be used to infer
object locations. 

"""


@torch.jit.script
def get_vector_mesh(shape: Tuple[int, int, int, int, int], device: str) -> Tensor:
    """generates a 3d mesh from a vector"""
    axis_ind: List[Tensor] = [
        torch.linspace(0, shape[2] - 1, shape[2], device=device),
        torch.linspace(0, shape[3] - 1, shape[3], device=device),
        torch.linspace(0, shape[4] - 1, shape[4], device=device),
    ]

    mesh: List[Tensor] = torch.meshgrid(axis_ind, indexing="ij")
    mesh: List[Tensor] = [m.unsqueeze(0).unsqueeze(0) for m in mesh]
    mesh: Tensor = torch.cat(mesh, dim=1)

    return mesh


@torch.jit.script
def _vec2embed3D_graphable(static_scale: Tensor, vector: Tensor, static_mesh) -> Tensor:
    """
    3D vector to embedding which uses static inputs for cuda graphs

    Could be a faster way to do this with strides but idk...

    :param scale: Tensor with shape (3)
    :param vector: [B, C, X, Y, Z]
    :return: embedding vector
    """

    return static_mesh + vector.mul(static_scale.view((1, 3, 1, 1, 1)))


@torch.jit.script
def _vec2embed2D(scale: Tensor, vector: Tensor) -> Tensor:
    """
    2D vector to embedding

    :param scale: The offest in XY of the vectors.
    :param vector: [B, C=2, X, Y] the vector matrix predicted by the unet

    :return: Pixel Spatial Embeddings (i.e. vector + pixel_indicies)
    """

    num: Tensor = torch.clone(scale.float())

    newshape: Tuple[int, int, int, int] = (1, 2, 1, 1)

    axis_ind: List[Tensor] = [
        torch.linspace(0, vector.shape[2] - 1, vector.shape[2], device=vector.device),
        torch.linspace(0, vector.shape[3] - 1, vector.shape[3], device=vector.device),
    ]

    mesh = torch.meshgrid(axis_ind, indexing="ij")
    mesh: List[Tensor] = [m.unsqueeze(0).unsqueeze(0) for m in mesh]
    mesh: Tensor = torch.cat(mesh, dim=1)

    vector = vector.mul(num.view(newshape))

    return mesh + vector


@torch.jit.script
def _vec2embed3D(scale: Tensor, vector: Tensor, n: int = 1) -> Tensor:
    """
    2D or 3D vector to embedding
    Could be a faster way to do this with strides but idk...
    :param scale: [N=2/3]
    :param vector: [B, C, X, Y, Z?]
    :param n: number of times to apply vectors
    :return:
    """

    num: Tensor = torch.clone(scale.float())

    newshape: Tuple[int, int, int, int, int] = (1, 3, 1, 1, 1)

    axis_ind: List[Tensor] = [
        torch.linspace(0, vector.shape[2] - 1, vector.shape[2], device=vector.device),
        torch.linspace(0, vector.shape[3] - 1, vector.shape[3], device=vector.device),
        torch.linspace(0, vector.shape[4] - 1, vector.shape[4], device=vector.device),
    ]

    mesh: List[Tensor] = torch.meshgrid(axis_ind, indexing="ij")
    mesh: List[Tensor] = [m.unsqueeze(0).unsqueeze(0) for m in mesh]
    mesh: Tensor = torch.cat(mesh, dim=1)

    scaled_vector = vector.mul(num.view(newshape))
    mesh = mesh + scaled_vector

    scale = 1.0
    for _ in range(n - 1):  # Only executes if n > 1
        # convert to index.

        scale *= 1.0

        scaled_vector = vector.mul(scale * num.view(newshape))

        index = mesh.round()
        b, c, x, y, z = index.shape
        for i, k in enumerate([x, y, z]):
            index[:, i, ...] = torch.clamp(index[:, i, ...], 0, k)

        # 3d index to raveled
        index = (
            (index[:, [0], ...] * y * z)
            + (index[:, [1], ...] * z)
            + (index[:, [2], ...])
        )
        index = index.clamp(0, x * y * z - 1).long()

        for i in range(c):
            mesh[:, [i], ...] = mesh[:, [i], ...] + scaled_vector[:, [i], ...].take(index)

    return mesh


def vector_to_embedding(scale: Tensor, vector: Tensor, N: int = 1) -> Tensor:
    """
    Converts a 2D or 3D vector field to a spatial embedding by adding the vector at any position to its own position.

    vector is a 2D or 3D vector field of shape :math:`(B, 2, X, Y)` for 2D or :math:`(B, 3, X, Y, Z)` for 3D.
    Each vector ":math:`v`" lies within the range -1 and 1 and is scaled by scale ":math:`s`". The scaled vector is then
    added to its own position to form a spatial embedding ":math:`\phi`":

    Formally:
        .. math::
            i,j,k \in \mathbb{Z}_{≥0} \n
            v_{i,j,k} \in [-1, 1] \n
            s = [s_i, s_j, s_k]

            \phi_{i,j,k} = v_{i,j,k} * s + [i, j, k]


    Shapes:
        - scale: :math:`(2)` or :math:`(3)`
        - vector: :math:`(B_{in}, 2, X_{in}, Y_{in})` or :math:`(B_{in}, 3, X_{in}, Y_{in}, Z_{in})`
        - Returns: :math:`(B_{in}, 2, X_{in}, Y_{in})` or :math:`(B_{in}, 3, X_{in}, Y_{in}, Z_{in})`


    :param scale: Scaling factors for each vector spatial dimension
    :param vector: Vector field predicted by a neural network

    :return: Pixel spatial embeddings
    """

    # assert vector.ndim in [4, 5], f'Vector must be a 4D or 5D tensor, not {vector.ndim}D: {vector.shape=}'
    return (
        _vec2embed3D(scale, vector, N)
        if vector.ndim == 5
        else _vec2embed2D(scale, vector)
    )


def vec2embedND(scale, vector):
    """
    Generic N dimmensional vector to embedding

    Could be a faster way to do this with strides but idk...

    :param scale: [N=2/3]
    :param vector: [B, C, X, Y, Z?]
    :return:
    """
    assert (
        scale.shape[0] == vector.shape[1]
    ), f"Cannot use {scale.shape[0]}D scale with vector shape: {vector.shape}"
    assert (
        scale.shape[0] == vector.ndim - 2
    ), f"Cannot use {scale.shape[0]}D scale with {vector.ndim - 2}D vector shape [B, C, ...]: {vector.shape}"

    num: Tensor = torch.clone(scale.float())

    newshape: Tuple[int] = tuple(
        [1, scale.shape[0]]
        + [
            1,
        ]
        * (vector.ndim - 2)
    )

    axis_ind: List[Tensor] = []
    for i in range(vector.ndim - 2):
        axis_ind.append(
            torch.linspace(
                0, vector.shape[2 + i] - 1, vector.shape[2 + i], device=vector.device
            )
        )

    mesh = torch.meshgrid(axis_ind, indexing="ij")
    mesh = [m.unsqueeze(0).unsqueeze(0) for m in mesh]
    mesh = torch.cat(mesh, dim=1)

    vector = vector.mul(num.view(newshape))

    return mesh + vector


if __name__ == "__main__":
    vector = torch.ones((1, 3, 10, 10, 10)).float()
    vector[:, 0, 5, 5, 5] = -1
    vector[:, 1, 5, 5, 5] = -1
    vector[:, 2, 5, 5, 5] = -1

    vector[:, [0,1,2], 4, 4, 4] = torch.tensor((2, 2, 2)).float()


    out = vector_to_embedding(torch.tensor((1, 1, 1)), vector, N=2)

    print(f"{out[0, :, 5, 5, 5]=}") # should equal 6 6 6



from __future__ import annotations
import glob
import os.path
from typing import Dict
from typing import Tuple, Callable, List, Union, Optional

import numpy as np
import skimage.io as io
import torch
from torch import Tensor
from torch.utils.data import Dataset
import logging

# from skoots.train.merged_transform import get_centroids
from tqdm import tqdm
from skoots.lib.custom_types import DataDict

Transform = Callable[[Dict[str, Tensor]], Dict[str, Tensor]]


class dataset(Dataset):
    def __init__(
        self,
        path: Union[List[str], str],
        transforms: Optional[Transform] = lambda x: x,
        pad_size: Optional[int] = 100,
        device: Optional[str] = "cpu",
        sample_per_image: Optional[int] = 1,
    ):
        r"""
        Custom dataset for loading and accessing skoots training data. This class loads data based on filenames and
        specific extensions: '.tif' (raw image), '.labels.tif' (instance masks), '.skeletons.tif' (precomputed skeletons).
        An example training data folder might contain the following:
        ::
            data\
             └  train\
                  │ train_data.tif
                  │ train_data.labels.tif
                  └ train_data.skeletons.tif


        :param path: Path to training data
        :param transforms: A function which applies dataset augmentation on a data_dict
        :param pad_size: padding to add to every image in the dataset
        :param device: torch.device which to **output** all data on
        :param sample_per_image: number of times each image/mask pair is sampled per iteration over a dataset
        """

        super(Dataset, self).__init__()

        # Reassigning variables
        self.files: List[str] = []
        self.image: List[Tensor] = []
        self.centroids: List[Tensor] = []
        self.masks: List[Tensor] = []
        self.skeletons: List[Dict[int, Tensor]] = []
        self.baked_skeleton: List[Tensor] = []
        self.transforms: Callable[[DataDict], DataDict] = transforms
        self.device = device
        self.pad_size: List[int] = [pad_size, pad_size]

        self.sample_per_image: int = sample_per_image

        # Store all possible directories containing data in a list
        path: List[str] = [path] if isinstance(path, str) else path

        for p in path:
            self.files.extend(glob.glob(f"{p}{os.sep}*.labels.tif"))

        for f in tqdm(self.files, desc="Loading Files: "):
            if os.path.exists(f[:-11:] + ".tif"):
                image_path = f[:-11:] + ".tif"
                logging.info(f'Loading Image: {image_path}')
            else:
                raise FileNotFoundError(
                    f"Could not find file: {image_path[:-4:]} with extensions .tif"
                )

            skeleton = (
                torch.load(f[:-11:] + ".skeletons.trch")
                if os.path.exists(f[:-11:] + ".skeletons.trch")
                else {-1: torch.tensor([])}
            )

            image: np.array = io.imread(image_path).astype(np.uint8)  # [Z, X, Y, C]
            masks: np.array = io.imread(f)  # [Z, X, Y]

            image: np.array = image[..., np.newaxis] if image.ndim == 3 else image
            image: np.array = image.transpose(-1, 1, 2, 0)
            image: np.array = image[[2], ...] if image.shape[0] > 3 else image

            masks: np.array = masks.transpose(1, 2, 0).astype(np.int32)

            assert image.max() < 256, f'16bit images not supported {image.max()=}'

            # Convert to torch.tensor
            image: Tensor = torch.from_numpy(image)  # .to(self.device)
            masks: Tensor = (
                torch.from_numpy(masks).int().unsqueeze(0)
            )  # .to(self.device)

            # I need the images in a float, but use torch automated mixed precision so can store as half.
            # This may not be the same for you!
            self.image.append(image)
            self.masks.append(masks)

            for i, (k, v) in enumerate(skeleton.items()):
                if v.numel() == 0:
                    raise ValueError(f"{f} instance label {k} has {v.numel()=}")

            self.skeletons.append(skeleton)
            self.baked_skeleton.append(None)


    def __len__(self) -> int:
        return len(self.image) * self.sample_per_image

    def __getitem__(self, item: int) -> DataDict:
        # We might artificially want to sample more times per image
        # Usefull when larging super large images with a lot of data.
        item = item // self.sample_per_image

        with torch.no_grad():
            data_dict: DataDict = {
                "image": self.image[item],
                "masks": self.masks[item],
                "skeletons": self.skeletons[item],
                "baked_skeleton": self.baked_skeleton[item],
            }

            # Transformation pipeline
            with torch.no_grad():
                data_dict: DataDict = self.transforms(data_dict)  # Apply transforms

        for k in data_dict:
            if isinstance(data_dict[k], torch.Tensor):
                data_dict[k] = data_dict[k].to(self.device)
            elif isinstance(data_dict[k], dict):
                data_dict[k] = {
                    key: value.to(self.device) for (key, value) in data_dict[k].items()
                }

        return data_dict

    def to(self, device: str):
        """
        Sends all data stored in the dataloader to a device.

        :param device: torch device for images, masks, and skeletons
        :return: self
        """
        self.image = [x.to(device) for x in self.image]
        self.masks = [x.to(device) for x in self.masks]
        self.skeletons = [
            {k: v.to(device) for (k, v) in x.items()} for x in self.skeletons
        ]

        return self

    def cuda(self) -> dataset:
        """alias for self.to('cuda:0')"""
        self.to("cuda:0")
        return self

    def cpu(self) -> dataset:
        """alias for self.to('cpu')"""
        self.to("cuda:0")
        self.to("cpu")
        return self

    def pin_memory(self) -> dataset:
        """
        Pins underlying memory allowing faster transfer to GPU
        """
        self.image = [x.pin_memory() for x in self.image]
        self.masks = [x.pin_memory() for x in self.masks]
        self.skeletons = [
            {k: v.pin_memory() for (k, v) in x.items()} for x in self.skeletons
        ]
        return self


class BackgroundDataset(Dataset):
    def __init__(
        self,
        path: Union[List[str], str],
        transforms: Optional[Transform] = lambda x: x,
        device: Optional[str] = "cpu",
        sample_per_image: Optional[int] = 1,
    ):
        super(Dataset, self).__init__()
        r"""
        Custom dataset for loading and accessing skoots background training data. 
        Unlike skoots.train.dataloader.dataset, which looks for masks and skeletons, 
        this dataset meed only images given that the images do not contain any actuall instances of the thing you're
        trying to segment - i.e. its background. 
        
        An example training data folder might contain the following:
        ::
            data\
             └  background\
                  └ background_image.tif


        :param path: Path to background data
        :param transforms: A function which applies background_dataset augmentation on a data_dict
        :param pad_size: padding to add to every image in the dataset
        :param device: torch.device which to **output** all data on
        :param sample_per_image: number of times each image/mask pair is sampled per iteration over a dataset
           """

        # Reassigning variables
        self.files = []
        self.image = []
        self.transforms = transforms
        self.device = device
        self.sample_per_image = sample_per_image

        path: List[str] = [path] if isinstance(path, str) else path

        for p in path:
            self.files.extend(glob.glob(f"{p}{os.sep}*.labels.tif"))

        for f in tqdm(self.files, desc="Loading Files: "):
            if os.path.exists(f[:-11:] + ".tif"):
                image_path = f[:-11:] + ".tif"
            else:
                raise FileNotFoundError(
                    f"Could not find file: {image_path[:-4:]} with extensions .tif"
                )

            skeleton = (
                torch.load(f[:-11:] + ".skeletons.trch")
                if os.path.exists(f[:-11:] + ".skeletons.trch")
                else {-1: torch.tensor([])}
            )

            image: np.array = io.imread(image_path)  # [Z, X, Y, C]
            masks: np.array = io.imread(f)  # [Z, X, Y]

            image: np.array = image[..., np.newaxis] if image.ndim == 3 else image
            image: np.array = image.transpose(-1, 1, 2, 0)
            image: np.array = image[[2], ...] if image.shape[0] > 3 else image

            scale: int = (
                2**16 if image.max() > 256 else 255
            )  # Our images might be 16 bit, or 8 bit
            scale: int = scale if image.max() > 1 else 1

            assert image.max() < 256, '16bit images not supported'
            image: Tensor = torch.from_numpy(image.astype(np.uint8))  # .to(self.device)
            self.image.append(image)

    def __len__(self) -> int:
        return len(self.image) * self.sample_per_image

    def __getitem__(self, item: int) -> Dict[str, Tensor]:
        # We might artificially want to sample more times per image
        # Usefull when larging super large images with a lot of data.
        item = item // self.sample_per_image

        with torch.no_grad():
            data_dict = {
                "image": self.image[item],
                "masks": torch.empty((1)),
                "skeletons": {-1: torch.empty((1))},
                "baked_skeleton": None,
            }

            # Transformation pipeline
            with torch.no_grad():
                data_dict = self.transforms(data_dict)  # Apply transforms

        for k in data_dict:
            if isinstance(data_dict[k], torch.Tensor):
                data_dict[k] = data_dict[k].to(self.device, non_blocking=True)
            elif isinstance(data_dict[k], dict):
                data_dict[k] = {
                    key: value.to(self.device, non_blocking=True)
                    for (key, value) in data_dict[k].items()
                }

        return data_dict

    def to(self, device: str):
        """
        Sends all data stored in the dataloader to a device.

        :param device: torch device for images, masks, and skeletons
        :return: self
        """
        self.image = [x.to(device) for x in self.image]
        self.masks = [x.to(device) for x in self.masks]
        self.skeletons = [
            {k: v.to(device) for (k, v) in x.items()} for x in self.skeletons
        ]

        return self

    def cuda(self):
        """alias for self.to('cuda:0')"""
        self.to("cuda:0")
        return self

    def cpu(self):
        """alias for self.to('cpu')"""
        self.to("cpu")
        return self


class MultiDataset(Dataset):
    def __init__(self, *args):
        r"""
        A utility class for joining multiple datasets into one accessible class. Sometimes, you may subdivide your
        training data based on some criteria. The most common is size: data from folder data/train/train_alot must be sampled 100 times
        per epoch, while data from folder data/train/train_notsomuch might only want to be sampled 1 times per epoch.

        You could construct a two skoots.train.dataloader.dataset objects for each
        and access both in a single MultiDataset class...

        >>> from skoots.train.dataloader import dataset
        >>>
        >>> # has one image sampled 100 times
        >>> data0 = dataset('data/train/train_alot', sample_per_image=100)
        >>> print(len(data0))  # 100
        >>>
        >>> # has one image sampled once
        >>> data1 = dataset('data/train/train_notsomuch', sample_per_image=1)
        >>> print(len(data1))  # 1
        >>>
        >>> merged_data = MultiDataset(data0, data1)
        >>> print(len(merged_data))  # 101, they've been merged!

        :param args:
        :type args:
        """
        self.datasets: List[Dataset] = []
        for ds in args:
            if isinstance(ds, Dataset):
                self.datasets.append(ds)

        self._dataset_lengths = [len(ds) for ds in self.datasets]

        self.num_datasets = len(self.datasets)

        self._mapped_indicies = []
        for i, ds in enumerate(self.datasets):
            # range(len(ds)) necessary to not index whole dataset at start. SLOW!!!
            self._mapped_indicies.extend([i for _ in range(len(ds))])

    def __len__(self) -> int:
        return len(self._mapped_indicies)

    def __getitem__(self, item: int) -> DataDict:
        i = self._mapped_indicies[item]  # Get the ind for the dataset
        _offset = sum(self._dataset_lengths[:i])  # Ind offset
        try:
            return self.datasets[i][item - _offset]
        except Exception as e:
            print(i, _offset, item - _offset, item, len(self.datasets[i]))
            raise e

    def to(self, device: str) -> MultiDataset:
        """
        Sends all data stored in the dataloader to a device. Occurs for ALL wrapped datasets.

        :param device: torch device for images, masks, and skeletons
        :return: self
        """
        for i in range(self.num_datasets):
            self.datasets[i].to(device)
        return self

    def cuda(self) -> MultiDataset:
        """alias for self.to('cuda:0')"""
        for i in range(self.num_datasets):
            self.datasets[i].to("cuda:0")
        return self

    def cpu(self) -> MultiDataset:
        """alias for self.to('cpu')"""
        for i in range(self.num_datasets):
            self.datasets[i].to("cpu")
        return self


# Custom batching function!
def skeleton_colate(
    data_dict: List[Dict[str, Tensor]]
) -> Tuple[Tensor, Tensor, List[Dict[str, Tensor]], Tensor, Tensor]:
    """
    Colate function with defines how we batch training data.
    Unpacks a data_dict with keys: 'image', 'masks', 'skele_masks', 'baked_skeleton', 'skeleton'
    and puts them each into a Tensor. This should not be called outright, rather passed to a
    torch.DataLoader for automatic batching.

    :param data_dict: Dictonary of augmented training data
    :return: Tuple of batched data
    """
    images = torch.stack([dd.pop("image") for dd in data_dict], dim=0)
    masks = torch.stack([dd.pop("masks") for dd in data_dict], dim=0)
    skele_masks = torch.stack([dd.pop("skele_masks") for dd in data_dict], dim=0)
    baked = [dd.pop("baked_skeleton") for dd in data_dict]

    if baked[0] is not None:
        baked = torch.stack(baked, dim=0)

    skeletons = [dd.pop("skeletons") for dd in data_dict]

    return images, masks, skeletons, skele_masks, baked


if __name__ == '__main__':
    """
    class dataset(Dataset):
    def __init__(
        self,
        path: Union[List[str], str],
        transforms: Optional[Transform] = lambda x: x,
        pad_size: Optional[int] = 100,
        device: Optional[str] = "cpu",
        sample_per_image: Optional[int] = 1,
    ):
    """
    from skoots.train.merged_transform import merged_transform_3D

    data = dataset(path='/home/chris/Dropbox (Partners HealthCare)/skoots-experiments/data/mitochondria/train/hide',
                   transforms=merged_transform_3D)

    for m in data.masks:
        print(m.max(), m.shape)

    for i in range(len(data)):
        print(data[i]['masks'].max(), data[i]['masks'].shape)





from typing import Dict, Optional, Union, List

import matplotlib.pyplot as plt
import torch
import torch.nn.modules.batchnorm
from torch import Tensor
from torchvision.utils import flow_to_image, draw_keypoints, make_grid


@torch.no_grad()
def update_bn(loader, model, device=None):
    r"""Updates BatchNorm running_mean, running_var buffers in the model.

    It performs one pass over data in `loader` to estimate the activation
    statistics for BatchNorm layers in the model.
    Args:
        loader (torch.utils.data.DataLoader): dataset loader to compute the
            activation statistics on. Each data batch should be either a
            tensor, or a list/tuple whose first element is a tensor
            containing data.
        model (torch.nn.Module): model for which we seek to update BatchNorm
            statistics.
        device (torch.device, optional): If set, data will be transferred to
            :attr:`device` before being passed into :attr:`model`.

    Example:
        >>> loader, model = ...
        >>> torch.optim.swa_utils.update_bn(loader, model)

    .. note::
        The `update_bn` utility assumes that each data batch in :attr:`loader`
        is either a tensor or a list or tuple of tensors; in the latter case it
        is assumed that :meth:`model.forward()` should be called on the first
        element of the list or tuple corresponding to the data batch.
    """
    momenta = {}
    for module in model.modules():
        if isinstance(module, torch.nn.modules.batchnorm._BatchNorm):
            module.running_mean = torch.zeros_like(module.running_mean)
            module.running_var = torch.ones_like(module.running_var)
            momenta[module] = module.momentum

    if not momenta:
        return

    was_training = model.training
    model.train()
    for module in momenta.keys():
        module.momentum = None
        module.num_batches_tracked *= 0

    for batch_ind, data_dict in enumerate(loader):
        images, data_dict = hcat.lib.utils.prep_dict(
            data_dict, device
        )  # Preps output to handle potential batched data
        output = model(images, data_dict)

    # for input in loader:
    #     if isinstance(input, (list, tuple)):
    #         input = input[0]
    #     if device is not None:
    #         input = input.to(device)
    #
    #     model(input)

    for bn_module in momenta.keys():
        bn_module.momentum = momenta[bn_module]
    model.train(was_training)


@torch.jit.script
def sum_loss(input: Dict[str, Tensor]) -> Optional[Tensor]:
    loss: Union[None, Tensor] = None
    for key in input:
        loss = loss + input[key] if loss is not None else input[key]
    return loss


device = "cuda:0" if torch.cuda.is_available() else "cpu"


def show_box_pred(image: List, output: List, thr=0.90):
    c = ["nul", "r", "b", "y", "w"]

    boxes = output[0]["boxes"].detach().cpu().numpy().tolist()
    labels = output[0]["labels"].detach().cpu().int().numpy().tolist()
    scores = (
        output[0]["scores"].detach().cpu().numpy().tolist()
        if "scores" in output[0]
        else [1] * len(labels)
    )

    image = image.cpu()

    # x1, y1, x2, y2
    inp = image.numpy().transpose((1, 2, 0))
    if inp.shape[-1] == 1:
        inp = inp[:, :, 0]
        plt.imshow(inp, origin="lower", cmap="Greys_r")
    else:
        plt.imshow(inp, origin="lower")

    for i, box in enumerate(boxes):
        if scores[i] < thr:
            continue

        x1 = box[0]
        y1 = box[1]
        x2 = box[2]
        y2 = box[3]

        plt.plot([x1, x2], [y2, y2], c[labels[i]], alpha=1 * scores[i])
        plt.plot([x1, x2], [y1, y1], c[labels[i]], alpha=1 * scores[i])
        plt.plot([x1, x1], [y1, y2], c[labels[i]], alpha=1 * scores[i])
        plt.plot([x2, x2], [y1, y2], c[labels[i]], alpha=1 * scores[i])

    return plt.gcf()


def mask_overlay(image: Tensor, mask: Tensor, thr: Optional[float] = 0.5) -> Tensor:
    if image.ndim > 3:
        raise RuntimeError("3D Tensors not supported!!!", image.shape)

    _image = image.gt(thr)
    _mask = mask.gt(thr)

    _, x, y = _image.shape

    overlap = _image * _mask
    false_positive = torch.logical_not(_image) * mask
    false_negative = torch.logical_not(_mask) * image

    out = torch.zeros((3, x, y), device=image.device)

    out[:, overlap[0, ...].gt(0.5)] = 1.0
    out[0, false_positive[0, ...].gt(0.5)] = 0.5
    out[2, false_negative[0, ...].gt(0.5)] = 0.5

    return out


def write_progress(
    writer,
    tag,
    epoch,
    images,
    masks,
    probability_map,
    vector,
    out,
    skeleton: Optional = None,
    predicted_skeleton: Optional = None,
    gt_skeleton: Optional = None,
):
    _a = images[0, [0, 0, 0], :, :, 7].cpu()
    _b = masks[0, [0, 0, 0], :, :, 7].gt(0.5).float().cpu()
    _overlay = mask_overlay(_b[[0], ...], probability_map[0, :, :, :, 7].cpu())
    _c = _overlay.mul(255).round().type(torch.uint8).cpu()

    if skeleton:
        for k in skeleton[0]:  # List[Dict[str, Tensor]]
            keypoints: Tensor = skeleton[0][k][:, [1, 0]].unsqueeze(1)
            _c = draw_keypoints(_c, keypoints, colors="blue", radius=1)

    _d = flow_to_image(vector[0, [1, 0], :, :, 7].float()).cpu()
    if skeleton:
        for k in skeleton[0]:  # List[Dict[str, Tensor]]
            keypoints: Tensor = skeleton[0][k][:, [1, 0]].unsqueeze(1)
            _d = draw_keypoints(_d, keypoints, colors="blue", radius=1)

    # print(out.shape)
    out: List[Tensor] = (
        [out[[0], ...]] if not isinstance(out, list) else out
    )  # [B, 1 ,x ,y,z] ?
    # print(type(out), out[0].shape)
    _f = out[0][0, ..., 7].max(0)[0].float().cpu().unsqueeze(0)
    # print(_f.shape)
    _f = torch.concat((_f, _f, _f), dim=0)
    img_list = [_a, _b, _c, _d, _f]

    if predicted_skeleton is not None:
        # print(predicted_skeleton.shape)
        _g = predicted_skeleton[0, 0, ..., 7].float().cpu().unsqueeze(0)
        _g = torch.concat((_g, _g, _g), dim=0).mul(255).round().type(torch.uint8)
        # print(_g.shape)
        assert _g.ndim == 3, f"{_g.shape=}"
        if skeleton is not None:
            for k in skeleton[0]:  # List[Dict[str, Tensor]]
                keypoints: Tensor = skeleton[0][k][:, [1, 0]].unsqueeze(1)
                _g = draw_keypoints(_g, keypoints, colors="blue", radius=1)
        img_list.append(_g)

    if gt_skeleton is not None:
        # print(predicted_skeleton.shape)
        _g = gt_skeleton[0, 0, ..., 7].float().cpu().unsqueeze(0)
        _g = torch.concat((_g, _g, _g), dim=0).mul(255).round().type(torch.uint8)
        # print(_g.shape)
        assert _g.ndim == 3, f"{_g.shape=}"
        img_list.append(_g)

    for i, im in enumerate(img_list):
        assert isinstance(
            im, torch.Tensor
        ), f"im {i} is not a Tensor instead it is a {type(im)}, {img_list[i]}"
        assert (
            img_list[0].shape == img_list[i].shape
        ), f"im {i} is has shape {im.shape}, not {img_list[0].shape}"

    _img = make_grid(img_list, nrow=1, normalize=True, scale_each=True)

    writer.add_image(tag, _img, epoch, dataformats="CWH")



from typing import Tuple, List, Optional

from torch import Tensor


def get_total_num_crops(
    image_shape: Tensor, crop_size: List[int], overlap: Optional[Tuple[int]]
) -> int:
    total = 0

    for i, size in enumerate(crop_size):
        crop_size[i] = (
            crop_size[i] if crop_size[i] < image_shape[i + 1] else image_shape[i + 1]
        )
        # overlap[i] = overlap[i] if cropsize[i] < image_shape[i+1] else 0

    assert len(image_shape) - 1 == len(crop_size) == len(overlap) == 3, (
        f"Image Shape must equal the shape of the crop.\n{image_shape=}, {crop_size=}"
        f"{overlap=}"
    )
    dim = ["x", "y", "z"]
    for c, o, d in zip(crop_size, overlap, dim):
        assert (
            c - o * 2 != 0
        ), f"Overlap in {d} dimmension cannot be equal to or larger than crop size... {o*2=} < {c}"

    x = 0
    while x < image_shape[1]:
        _x = x if x + crop_size[0] <= image_shape[1] else image_shape[1] - crop_size[0]

        y = 0
        while y < image_shape[2]:
            _y = (
                y
                if y + crop_size[1] <= image_shape[2]
                else image_shape[2] - crop_size[1]
            )

            z = 0
            while z < image_shape[3]:
                _z = (
                    z
                    if z + crop_size[2] <= image_shape[3]
                    else image_shape[3] - crop_size[2]
                )

                total += 1

                z += crop_size[2] - (overlap[2] * 2)
            y += crop_size[1] - (overlap[1] * 2)
        x += crop_size[0] - (overlap[0] * 2)

    return total


def crops(
    image: Tensor,
    crop_size: List[int],
    overlap: Optional[Tuple[int, int, int]] = (0, 0, 0),
    device="cpu",
) -> Tuple[Tensor, List[int]]:
    """
    Generator which takes an image and sends out crops of a certain size with overlap pixels
    Can send a crop to a torch.device ahead of use through caching.

    Shapes:
        - image: :math:`(C, X, Y, Z)`
        - yeilds: :math:`(B, C, X, Y, Z)`

    :param image: 4D torch.Tensor of shape [C, X, Y, Z]
    :param crop_size: Spatial dims of the resulting crops [X, Y, Z]
    :param overlap: Overlap between each crop
    :param device: torch device of the crop
    :return: Crop of the image, and the indicies of the crop
    """

    image_shape = image.shape  # C, X, Y, Z

    for i, size in enumerate(crop_size):
        crop_size[i] = (
            crop_size[i] if crop_size[i] < image_shape[i + 1] else image_shape[i + 1]
        )

    assert len(image_shape) - 1 == len(crop_size) == len(overlap) == 3, (
        f"Image Shape must equal the shape of the crop.\n{image.shape=}, {crop_size=}"
        f"{overlap=}"
    )
    dim = ["x", "y", "z"]
    for c, o, d in zip(crop_size, overlap, dim):
        assert (
            c - (o * 2) != 0
        ), f"Overlap in {d} dimmension cannot be equal to or larger than crop size... {c=} - {o*2=} = {c - (o * 2)} < {c}"

    # for i in range(image_shape[1] // cropsize[1] + 1):

    _output_cache = []

    x = 0
    while x < image_shape[1]:
        _x = x if x + crop_size[0] <= image_shape[1] else image_shape[1] - crop_size[0]

        y = 0
        while y < image_shape[2]:
            _y = (
                y
                if y + crop_size[1] <= image_shape[2]
                else image_shape[2] - crop_size[1]
            )

            z = 0
            while z < image_shape[3]:
                _z = (
                    z
                    if z + crop_size[2] <= image_shape[3]
                    else image_shape[3] - crop_size[2]
                )

                _output_cache.append(
                    (
                        image[
                            :,
                            _x : _x + crop_size[0],
                            _y : _y + crop_size[1],
                            _z : _z + crop_size[2],
                        ]
                        .unsqueeze(0)
                        .to(device, non_blocking=True),
                        [_x, _y, _z],
                    )
                )

                if len(_output_cache) > 1:
                    yield _output_cache.pop(0)

                z += crop_size[2] - (overlap[2] * 2)
            y += crop_size[1] - (overlap[1] * 2)
        x += crop_size[0] - (overlap[0] * 2)

    yield _output_cache.pop(0)  # final yield



from yacs.config import CfgNode as CN

# -----------------------------------------------------------------------------
# Training config definition
# -----------------------------------------------------------------------------
_C = CN()

# -----------------------------------------------------------------------------
# System
# -----------------------------------------------------------------------------
_C.SYSTEM = CN()

_C.SYSTEM.NUM_GPUS = 2
_C.SYSTEM.NUM_CPUS = 1

# Define a BISM Model
_C.MODEL = CN()
_C.MODEL.ARCHITECTURE = "bism_unext"
_C.MODEL.IN_CHANNELS = 1
_C.MODEL.OUT_CHANNELS = 32
_C.MODEL.DIMS = [32, 64, 128, 64, 32]
_C.MODEL.DEPTHS = [2, 2, 2, 2, 2]
_C.MODEL.KERNEL_SIZE = 7
_C.MODEL.DROP_PATH_RATE = 0.0
_C.MODEL.LAYER_SCALE_INIT_VALUE = 1.0
_C.MODEL.ACTIVATION = "gelu"
_C.MODEL.BLOCK = "block3d"
_C.MODEL.CONCAT_BLOCK = "concatconv3d"
_C.MODEL.UPSAMPLE_BLOCK = "upsamplelayer3d"
_C.MODEL.NORMALIZATION = "layernorm"

# Training Configurations
_C.TRAIN = CN()
_C.TRAIN.DISTRIBUTED = True
_C.TRAIN.PRETRAINED_MODEL_PATH = [
    "/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/models/Oct20_11-54-51_CHRISUBUNTU.trch"
]

# Loss function and their constructor keyowrds
_C.TRAIN.LOSS_EMBED = "tversky"
_C.TRAIN.LOSS_EMBED_KEYWORDS = ["alpha", "beta", "eps"]
_C.TRAIN.LOSS_EMBED_VALUES = [0.25, 0.75, 1e-8]

_C.TRAIN.LOSS_PROBABILITY = "tversky"
_C.TRAIN.LOSS_PROBABILITY_KEYWORDS = ["alpha", "beta", "eps"]
_C.TRAIN.LOSS_PROBABILITY_VALUES = [0.5, 0.5, 1e-8]

_C.TRAIN.LOSS_SKELETON = "tversky"
_C.TRAIN.LOSS_SKELETON_KEYWORDS = ["alpha", "beta", "eps"]
_C.TRAIN.LOSS_SKELETON_VALUES = [0.5, 1.5, 1e-8]

# We sum the loss values of each part together, scaled by some factor set here
_C.TRAIN.LOSS_EMBED_RELATIVE_WEIGHT = 1.0
_C.TRAIN.LOSS_PROBABILITY_RELATIVE_WEIGHT = 1.0
_C.TRAIN.LOSS_SKELETON_RELATIVE_WEIGHT = 1.0

# We may not consider each loss until a certain epoch. This may be
# because some tasks are hard to learn at the start, and must only be considered later
_C.TRAIN.LOSS_EMBED_START_EPOCH = -1
_C.TRAIN.LOSS_PROBABILITY_START_EPOCH = -1
_C.TRAIN.LOSS_SKELETON_START_EPOCH = 10

_C.TRAIN.TRAIN_DATA_DIR = []
_C.TRAIN.TRAIN_SAMPLE_PER_IMAGE = []
_C.TRAIN.TRAIN_BATCH_SIZE = 1
_C.TRAIN.VALIDATION_DATA_DIR = []
_C.TRAIN.VALIDATION_SAMPLE_PER_IMAGE = []
_C.TRAIN.VALIDATION_BATCH_SIZE = 1

_C.TRAIN.BACKGROUND_DATA_DIR = []
_C.TRAIN.BACKGROUND_SAMPLE_PER_IMAGE = []

_C.TRAIN.TRAIN_STORE_DATA_ON_GPU = []
_C.TRAIN.VALIDATION_STORE_DATA_ON_GPU = []
_C.TRAIN.BACKGROUND_STORE_DATA_ON_GPU = []

_C.TRAIN.STORE_DATA_ON_GPU = []
_C.TRAIN.INITIAL_SIGMA = [20.0, 20.0, 20.0]
_C.TRAIN.SIGMA_DECAY = [
    [0.66, 200],
    [0.66, 800],
    [0.66, 1500],
    [0.5, 20000],
    [0.5, 20000],
]
_C.TRAIN.NUM_EPOCHS = 10000
_C.TRAIN.LEARNING_RATE = 5e-4
_C.TRAIN.WEIGHT_DECAY = 1e-6
_C.TRAIN.OPTIMIZER = "adamw"  # Adam, AdamW, SGD,
_C.TRAIN.OPTIMIZER_KEYWORD_ARGUMENTS = []
_C.TRAIN.OPTIMIZER_KEYWORD_VALUES = []
_C.TRAIN.OPTIMIZER_EPS = 1e-8
_C.TRAIN.SCHEDULER = "cosine_annealing_warm_restarts"
_C.TRAIN.SCHEDULER_T0 = 10000 + 1
_C.TRAIN.MIXED_PRECISION = True
_C.TRAIN.N_WARMUP = 1500
_C.TRAIN.SAVE_PATH = (
    "/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/models"
)
_C.TRAIN.SAVE_INTERVAL = 100
_C.TRAIN.CUDNN_BENCHMARK = True
_C.TRAIN.AUTOGRAD_PROFILE = False
_C.TRAIN.AUTOGRAD_EMIT_NVTX = False
_C.TRAIN.AUTOGRAD_DETECT_ANOMALY = False

# Augmentation
_C.AUGMENTATION = CN()
_C.AUGMENTATION.CROP_WIDTH = 300
_C.AUGMENTATION.CROP_HEIGHT = 300
_C.AUGMENTATION.CROP_DEPTH = 20
_C.AUGMENTATION.FLIP_RATE = 0.5
_C.AUGMENTATION.BRIGHTNESS_RATE = 0.4
_C.AUGMENTATION.BRIGHTNESS_RANGE = [-0.1, 0.1]
_C.AUGMENTATION.NOISE_GAMMA = 0.1
_C.AUGMENTATION.NOISE_RATE = 0.2
_C.AUGMENTATION.CONTRAST_RATE = 0.33
_C.AUGMENTATION.CONTRAST_RANGE = [0.75, 2.0]
_C.AUGMENTATION.AFFINE_RATE = 0.66
_C.AUGMENTATION.AFFINE_SCALE = [0.85, 1.1]
_C.AUGMENTATION.AFFINE_YAW = [-180, 180]
_C.AUGMENTATION.AFFINE_SHEAR = [-7, 7]
_C.AUGMENTATION.SMOOTH_SKELETON_KERNEL_SIZE = (3, 3, 1)
_C.AUGMENTATION.BAKE_SKELETON_ANISOTROPY = (1.0, 1.0, 3.0)
_C.AUGMENTATION.N_SKELETON_MASK_DILATE = 1

# Skoots Generics
_C.SKOOTS = CN()
_C.SKOOTS.VECTOR_SCALING = (60, 60, 60 // 5)
_C.SKOOTS.ANISOTROPY = (1.0, 1.0, 3.0)
_C.SKOOTS.NOTES = ''


def get_cfg_defaults():
    r"""Get a yacs CfgNode object with default values for my_project."""
    # Return a clone so that the defaults will not be altered
    # This is for the "local variable" use pattern
    return _C.clone()



import torch
from torch import Tensor


@torch.jit.script
def baked_embed_to_prob(
    embedding: Tensor, baked_skeletons: Tensor, sigma: Tensor, eps: float = 1e-16
) -> Tensor:
    r"""
    N Dimensional embedding to probability with a baked skeleton array

    Calculates a probability :math:`\phi` based on a euclidean distance between a spatial
    embedding :math:`E_i` and a baked skeleton pixel :math:`S_i`.

    .. math::
        \phi(E_i, S_i) =exp\left(\sum_{k \in [x,y,z]} \frac{(E_{ki} - S_{ki})^2}{-2\sigma^2_k} \right)

    In three spatial dimmensions, this expands to

    .. math::
        \phi(E_i, S_i) =exp\left(\frac{(E_{xi} - S_{xi})^2}{-2\sigma^2_x} + \frac{(E_{yi} - S_{yi})^2}{-2\sigma_y^2} + \frac{(E_{zi} - S_{zi})^2}{-2\sigma^2_z}\right)

    Shapes:
        - embedding: :math:`(B_{in}, 2, X_{in}, Y_{in})` or :math:`(B_{in}, 3, X_{in}, Y_{in}, Z_{in})`
        - baked_skeletons: :math:`(B_{in}, 2, X_in, Y_{in})` or :math:`(B_{in}, 3, X_{in}, Y_in, Z_{in})`
        - sigma: :math:`(2)` or :math:`(3)`

        - returns: :math:`(B_{in}, 1, X_{in}, Y_{in})` or :math:`(B_{in}, 1, X_{in}, Y_{in}, Z_{in})`

    :param embedding: embedding tensor
    :param baked_skeletons: a baked skeleton tensor
    :param sigma: Standard deviation of the gaussian. Larger values give higher probability further away.
    :param eps: small value for numerical stability
    :return: Probability matrix
    """

    sigma = sigma + eps  # when sigma goes to zero, things tend to break
    sigma = sigma.pow(2).mul(2).mul(-1)

    out = torch.exp(
        (embedding - baked_skeletons)
        .pow(2)
        .transpose(
            1, -1
        )  # work for 2D and 3D by following pytorch broadcasting rules (channels last dim)
        .div(sigma)
        .transpose(1, -1)
        .sum(dim=1, keepdim=True)
    )

    return out


if __name__ == "__main__":
    pass



import os.path
import warnings

import numpy as np
import skimage.io as io
import torch
import torch.nn.functional as F
from bism.models import get_constructor
from bism.models.spatial_embedding import SpatialEmbedding
from torch import Tensor
from torch.cuda.amp import autocast
from tqdm import tqdm

import skoots.lib.skeleton
from skoots.lib.cropper import crops
from skoots.lib.flood_fill import efficient_flood_fill
from skoots.lib.vector_to_embedding import vector_to_embedding

warnings.filterwarnings("ignore")

"""
Instance Segmentation more or less...
---------------------
    vectors, skeletons, semantic_masks = model(image)
    embedding_x, embedding_y, embedding_z = vectors_to_embedding(vectors * semantic_masks)
    instance_skeletons = efficient_flood_fill(skeletons)
    instance_mask = instance_skeleton[embedding_x, embedding_y, embedding_z]

"""


@torch.inference_mode()  # disables autograd and reference counting for SPEED
def eval(
    image_path: str,
    num_tuple=(60, 60, 60 // 5),
    checkpoint_path: str = "/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/models/Oct21_17-15-08_CHRISUBUNTU.trch",
) -> None:
    scale = -99999

    image: np.array = io.imread(image_path)  # [Z, X, Y, C]
    image: np.array = image[..., np.newaxis] if image.ndim == 3 else image
    image: np.array = image.transpose(-1, 1, 2, 0)
    image: np.array = image[[2], ...] if image.shape[0] > 3 else image  # [C=1, X, Y, Z]

    scale: int = 2**16 if image.dtype == np.uint16 else 2**8

    image: Tensor = torch.from_numpy(image.astype(np.int32)).pin_memory()
    print(f"Image Shape: {image.shape}, Dtype: {image.dtype}, Scale Factor: {scale}")

    # Allocate a bunch or things...
    if image.dtype == torch.float:
        pad3d = (5, 5, 30, 30, 30, 30)  # Pads last dim first!
        image = (
            F.pad(image, pad3d, mode="reflect") if image.dtype == torch.float else image
        )
    else:
        pad3d = False

    # num_tuple = (60, 60, 60 // 5)
    num = torch.tensor(num_tuple)

    c, x, y, z = image.shape

    skeleton = torch.zeros(size=(1, x, y, z), dtype=torch.int16)
    # semantic = torch.zeros(size=(1, x, y, z), dtype=torch.int16)
    vectors = torch.zeros((3, x, y, z), dtype=torch.half)

    device = "cuda:0" if torch.cuda.is_available() else "cpu"

    checkpoint = torch.load(checkpoint_path)

    state_dict = (
        checkpoint
        if not "model_state_dict" in checkpoint
        else checkpoint["model_state_dict"]
    )

    model_constructor = get_constructor(
        "unext", spatial_dim=3
    )  # gets the model from a name...
    backbone = model_constructor(
        in_channels=1, out_channels=16, dims=[16, 32, 64, 32, 16]
    )
    model = SpatialEmbedding(backbone=backbone)
    model.load_state_dict(state_dict=state_dict)
    model = model.to(device).train()
    # model = torch.jit.optimize_for_inference(torch.jit.script(model))

    cropsize = [300, 300, 20]
    overlap = [30, 30, 2]

    total = skoots.lib.cropper.get_total_num_crops(image.shape, cropsize, overlap)
    iterator = tqdm(crops(image, cropsize, overlap), desc="", total=total)

    id = 1
    for slice, (x, y, z) in iterator:
        with autocast(enabled=True):  # Saves Memory!
            out = model(slice.div(scale).float().cuda())

        probability_map = out[:, [-1], ...].cpu()
        skeleton_map = out[:, [-2], ...].float()

        skeleton_map = skeleton_map.cpu()
        vec = out[:, 0:3:1, ...].cpu()

        vec = vec * probability_map.gt(0.5)
        skeleton_map = skeleton_map * probability_map.gt(0.5)

        skeleton[
            :,
            x + overlap[0] : x + cropsize[0] - overlap[0],
            y + overlap[1] : y + cropsize[1] - overlap[1],
            z + overlap[2] : z + cropsize[2] - overlap[2],
        ] = skeleton_map[
            0,
            :,
            overlap[0] : -overlap[0],
            overlap[1] : -overlap[1],
            overlap[2] : -overlap[2] :,
        ].gt(
            0.8
        )
        # semantic[:,
        # x + overlap[0]: x + cropsize[0] - overlap[0],
        # y + overlap[1]: y + cropsize[1] - overlap[1],
        # z + overlap[2]: z + cropsize[2] - overlap[2]] = probability_map[0, :,
        #                                                 overlap[0]: -overlap[0],
        #                                                 overlap[1]: -overlap[1],
        #                                                 overlap[2]: -overlap[2]
        #                                                 :].gt(0.5)

        vectors[
            :,
            x + overlap[0] : x + cropsize[0] - overlap[0],
            y + overlap[1] : y + cropsize[1] - overlap[1],
            z + overlap[2] : z + cropsize[2] - overlap[2],
        ] = vec[
            0,
            :,
            overlap[0] : -overlap[0],
            overlap[1] : -overlap[1],
            overlap[2] : -overlap[2] :,
        ].half()

        iterator.desc = f"Evaluating UNet on slice [x{x}:y{y}:z{z}]"

    # _x, _y, _z
    if pad3d:
        skeleton = skeleton[
            0, pad3d[2] : -pad3d[3], pad3d[4] : -pad3d[5], pad3d[0] : -pad3d[1]
        ]
        # semantic = semantic[0, pad3d[2]:-pad3d[3], pad3d[4]:-pad3d[5], pad3d[0]:-pad3d[1]]
        vectors = vectors[
            :, pad3d[2] : -pad3d[3], pad3d[4] : -pad3d[5], pad3d[0] : -pad3d[1]
        ]

    else:
        skeleton = skeleton[0, ...]
        vectors = vectors[:, ...]

    del image

    # io.imsave('/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/outputs/semantic.tif',
    #           semantic.squeeze().mul(255).int().cpu().numpy().astype(np.uint8).transpose(2, 0, 1))

    # io.imsave is memory costly when saving large tensors... :(
    torch.save(vectors, os.path.split(image_path)[0] + "/vectors.trch")
    torch.save(skeleton, os.path.split(image_path)[0] + "/skeleton_unlabeled.trch")

    # io.imsave(
    #     '/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/outputs/skeleton_unlabeled.tif',
    #     skeleton.squeeze().cpu().numpy().transpose(2, 0, 1))

    skeleton = efficient_flood_fill(skeleton)

    # io.imsave('/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/outputs/skeleton.tif',
    #           skeleton.cpu().numpy().transpose(2, 0, 1))

    torch.save(skeleton, os.path.split(image_path)[0] + "/skeleton.trch")

    instance_mask = torch.zeros_like(skeleton, dtype=torch.int16)
    skeleton = skeleton.unsqueeze(0).unsqueeze(0)

    iterator = tqdm(
        crops(vectors, crop_size=[500, 500, 50]), desc="Assigning Instances:"
    )

    print(f"[      ] identifying connected components...", end="")
    for _vec, (x, y, z) in iterator:
        _embed = skoots.lib.vector_to_embedding.vector_to_embedding(
            scale=num, vector=_vec, N=1
        )
        _embed += torch.tensor((x, y, z)).view(
            1, 3, 1, 1, 1
        )  # We adjust embedding to region of the crop
        _inst_maks = skoots.lib.skeleton.index_skeleton_by_embed(
            skeleton=skeleton, embed=_embed
        ).squeeze()
        w, h, d = _inst_maks.shape
        instance_mask[x : x + w, y : y + h, z : z + d] = _inst_maks

    print("DONE")
    # print(instance_mask.unique().shape[0] - 1, ' Unique mito')

    del skeleton, vectors  # explicitly delete unnecessary tensors for memory

    io.imsave(
        os.path.split(image_path)[0] + "/instance_mask.tif",
        instance_mask.cpu().numpy().astype(np.uint16).transpose(2, 0, 1),
    )


if __name__ == "__main__":
    # image_path = '/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/outputs/hide_validate-1.tif'
    # image_path = '/home/chris/Documents/threeOHC_registered_8bit_cell2.tif'
    image_path = "/home/chris/Documents/threeOHC_registered_8bit.tif"
    # image_path = '/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/outputs/onemito.tif'
    # image_path = '/home/chris/Dropbox (Partners HealthCare)/trainMitochondriaSegmentation/outputs/cell_apex-1.tif'
    # image_path = '/home/chris/Dropbox (Partners HealthCare)/Manuscripts - Buswinka/Mitochondria Segmentation/Figures/Figure X6X  - Whole image analysis/crop.tif'
    eval(image_path)



from functools import partial
from numbers import Number
from typing import List, Tuple, Optional

import bism.backends
import bism.modules
import torch
import torch.nn as nn
from bism.models.spatial_embedding import SpatialEmbedding
from torch import Tensor
from yacs.config import CfgNode


def cfg_to_bism_model(cfg: CfgNode) -> nn.Module:
    """utility function to get a bism model from cfg"""

    _valid_model_constructors = {
        "bism_unext": bism.backends.unext.UNeXT_3D,
        "bism_unet": bism.backends.unet.UNet_3D,
    }

    _valid_model_blocks = {"block3d": bism.modules.convnext_block.Block3D}

    _valid_upsample_layers = {
        "upsamplelayer3d": bism.modules.upsample_layer.UpSampleLayer3D
    }

    _valid_normalization = {
        "layernorm": partial(
            bism.modules.layer_norm.LayerNorm, data_format="channels_first"
        )
    }

    _valid_activations = {
        "gelu": torch.nn.GELU,
        "relu": torch.nn.ReLU,
        "silu": torch.nn.SiLU,
        "selu": torch.nn.SELU,
    }

    _valid_concat_blocks = {"concatconv3d": bism.modules.concat.ConcatConv3D}

    model_config = [
        cfg.MODEL.DIMS,
        cfg.MODEL.DEPTHS,
        cfg.MODEL.KERNEL_SIZE,
        cfg.MODEL.DROP_PATH_RATE,
        cfg.MODEL.LAYER_SCALE_INIT_VALUE,
        cfg.MODEL.ACTIVATION,
        cfg.MODEL.BLOCK,
        cfg.MODEL.CONCAT_BLOCK,
        cfg.MODEL.UPSAMPLE_BLOCK,
        cfg.MODEL.NORMALIZATION,
    ]

    model_kwargs = [
        "dims",
        "depths",
        "kernel_size",
        "drop_path_rate",
        "layer_scale_init_value",
        "activation",
        "block",
        "concat_conv",
        "upsample_layer",
        "normalization",
    ]

    valid_dicts = [
        None,
        None,
        None,
        None,
        None,
        _valid_activations,
        _valid_model_blocks,
        _valid_concat_blocks,
        _valid_upsample_layers,
        _valid_normalization,
    ]

    kwarg = {}
    for kw, config, vd in zip(model_kwargs, model_config, valid_dicts):
        if vd is not None:
            if config in vd:
                kwarg[kw] = vd[config]
            else:
                raise RuntimeError(
                    f"{config} is not a valid config option for {kw}. Valid options are: {vd.keys()}"
                )
        else:
            kwarg[kw] = config

    if cfg.MODEL.ARCHITECTURE in _valid_model_constructors:
        backbone = _valid_model_constructors[cfg.MODEL.ARCHITECTURE]
    else:
        raise RuntimeError(
            f"{cfg.MODEL.ARCHITECTURE} is not a valid model constructor, valid options are: {_valid_model_constructors.keys()}"
        )

    backbone = backbone(cfg.MODEL.IN_CHANNELS, cfg.MODEL.OUT_CHANNELS, **kwarg)
    model = SpatialEmbedding(backbone)

    return model


def calculate_indexes(
    pad_size: int, eval_image_size: int, image_shape: int, padded_image_shape: int
) -> List[List[int]]:
    """
    This calculates indexes for the complete evaluation of an arbitrarily large image by unet.
    each index is offset by eval_image_size, but has a width of eval_image_size + pad_size * 2.
    Unet needs padding on each side of the evaluation to ensure only full convolutions are used
    in generation of the final mask. If the algorithm cannot evenly create indexes for
    padded_image_shape, an additional index is added at the end of equal size.



    :param pad_size: int corresponding to the amount of padding on each side of the
                     padded image
    :param eval_image_size: int corresponding to the shape of the image to be used for
                            the final mask
    :param image_shape: int Shape of image before padding is applied
    :param padded_image_shape: int Shape of image after padding is applied

    :return: List of lists corresponding to the indexes
    """

    # We want to account for when the eval image size is super big, just return index for the whole image.
    if eval_image_size + (2 * pad_size) > image_shape:
        return [[0, image_shape - 1]]

    try:
        ind_list = torch.arange(0, image_shape, eval_image_size)
    except RuntimeError:
        raise RuntimeError(
            f"Calculate_indexes has incorrect values {pad_size} | {image_shape} | {eval_image_size}:\n"
            f"You are likely trying to have a chunk smaller than the set evaluation image size. "
            "Please decrease number of chunks."
        )
    ind = []
    for i, z in enumerate(ind_list):
        if i == 0:
            continue
        z1 = int(ind_list[i - 1])
        z2 = int(z - 1) + (2 * pad_size)
        if z2 < padded_image_shape:
            ind.append([z1, z2])
        else:
            break
    if (
        not ind
    ):  # Sometimes z is so small the first part doesnt work. Check if z_ind is empty, if it is do this!!!
        z1 = 0
        z2 = eval_image_size + pad_size * 2
        ind.append([z1, z2])
        ind.append(
            [padded_image_shape - (eval_image_size + pad_size * 2), padded_image_shape]
        )
    else:  # we always add at the end to ensure that the whole thing is covered.
        z1 = padded_image_shape - (eval_image_size + pad_size * 2)
        z2 = padded_image_shape - 1
        ind.append([z1, z2])
    return ind


def get_dtype_offset(dtype: str = "uint16", image_max: Optional[Number] = None) -> int:
    """
    Returns the scaling factor such that
    such that :math:`\frac{image}{f} \in [0, ..., 1]`

    Supports: uint16, uint8, uint12, float64

    :param dtype: String representation of the data type.
    :param image_max: Returns image max if the dtype is not supported.
    :return: Integer scale factor
    """

    encoding = {
        "uint16": 2**16,
        "uint8": 2**8,
        "uint12": 2**12,
        "float64": 1,
    }
    dtype = str(dtype)
    if dtype in encoding:
        scale = encoding[dtype]
    else:
        print(
            f"\x1b[1;31;40m"
            + f"ERROR: Unsupported dtype: {dtype}. Currently support: {[k for k in encoding]}"
            + "\x1b[0m"
        )
        scale = None
        if image_max:
            print(f"Appropriate Scale Factor inferred from image maximum: {image_max}")
            if image_max <= 256:
                scale = 256
            else:
                scale = image_max
    return scale


@torch.jit.script
def _crop3d(img: Tensor, x: int, y: int, z: int, w: int, h: int, d: int) -> Tensor:
    """
    torch scriptable function which crops an image

    :param img: torch.Tensor image of shape [C, X, Y, Z]
    :param x: x coord of crop box
    :param y: y coord of crop box
    :param z: z coord of crop box
    :param w: width of crop box
    :param h: height of crop box
    :param d: depth of crop box
    :return:
    """
    return img[..., x : x + w, y : y + h, z : z + d]


@torch.jit.script
def _crop2d(img: Tensor, x: int, y: int, w: int, h: int) -> Tensor:
    """
    torch scriptable function which crops an image

    :param img: torch.Tensor image of shape [C, X, Y, Z]
    :param x: x coord of crop box
    :param y: y coord of crop box
    :param z: z coord of crop box
    :param w: width of crop box
    :param h: height of crop box
    :param d: depth of crop box
    :return:
    """

    return img[..., x : x + w, y : y + h]


@torch.jit.script
def crop_to_identical_size(a: Tensor, b: Tensor) -> Tuple[Tensor, Tensor]:
    """
    Crops Tensor a to the shape of Tensor b, then crops Tensor b to the shape of Tensor a.

    :param a: torch.
    :param b:
    :return:
    """
    if a.ndim < 3:
        raise RuntimeError(
            "Only supports tensors with minimum 3dimmensions and shape [..., X, Y, Z]"
        )

    a = _crop3d(a, x=0, y=0, z=0, w=b.shape[-3], h=b.shape[-2], d=b.shape[-1])
    b = _crop3d(b, x=0, y=0, z=0, w=a.shape[-3], h=a.shape[-2], d=a.shape[-1])
    return a, b


@torch.jit.script
def cantor2(a: Tensor, b: Tensor) -> Tensor:
    "Hashes two combination of tensors together"
    return 0.5 * (a + b) * (a + b + 1) + b


@torch.jit.script
def cantor3(a: Tensor, b: Tensor, c: Tensor) -> Tensor:
    "Hashes three combination of tensors together"
    return (0.5 * (cantor2(a, b) + c) * (cantor2(a, b) + c + 1) + c).int()


@torch.jit.script
def identical_rows(a: Tensor, b: Tensor) -> Tensor:
    """
    Given two matrices of identical size, determines the indices of identical rows.

    :param a: [N, 3] torch.Tensor
    :param b: [N, 3] torch.Tensor

    :return: Indicies of identical rows
    """
    a = cantor3(a[:, 0], a[:, 1], a[:, 2])
    b = cantor3(b[:, 0], b[:, 1], b[:, 2])

    _inv = torch.reciprocal(b)

    ind = torch.sum(a.unsqueeze(-1) @ _inv.unsqueeze(0), dim=0).gt(0)

    return ind


def get_cached_ball_coords(device: str) -> Tensor:
    out = [
        [0, 3, 3],
        [1, 1, 2],
        [1, 1, 3],
        [1, 1, 4],
        [1, 2, 1],
        [1, 2, 2],
        [1, 2, 3],
        [1, 2, 4],
        [1, 2, 5],
        [1, 3, 1],
        [1, 3, 2],
        [1, 3, 3],
        [1, 3, 4],
        [1, 3, 5],
        [1, 4, 1],
        [1, 4, 2],
        [1, 4, 3],
        [1, 4, 4],
        [1, 4, 5],
        [1, 5, 2],
        [1, 5, 3],
        [1, 5, 4],
        [2, 1, 1],
        [2, 1, 2],
        [2, 1, 3],
        [2, 1, 4],
        [2, 1, 5],
        [2, 2, 1],
        [2, 2, 2],
        [2, 2, 3],
        [2, 2, 4],
        [2, 2, 5],
        [2, 3, 1],
        [2, 3, 2],
        [2, 3, 3],
        [2, 3, 4],
        [2, 3, 5],
        [2, 4, 1],
        [2, 4, 2],
        [2, 4, 3],
        [2, 4, 4],
        [2, 4, 5],
        [2, 5, 1],
        [2, 5, 2],
        [2, 5, 3],
        [2, 5, 4],
        [2, 5, 5],
        [3, 0, 3],
        [3, 1, 1],
        [3, 1, 2],
        [3, 1, 3],
        [3, 1, 4],
        [3, 1, 5],
        [3, 2, 1],
        [3, 2, 2],
        [3, 2, 3],
        [3, 2, 4],
        [3, 2, 5],
        [3, 3, 0],
        [3, 3, 1],
        [3, 3, 2],
        [3, 3, 3],
        [3, 3, 4],
        [3, 3, 5],
        [3, 3, 6],
        [3, 4, 1],
        [3, 4, 2],
        [3, 4, 3],
        [3, 4, 4],
        [3, 4, 5],
        [3, 5, 1],
        [3, 5, 2],
        [3, 5, 3],
        [3, 5, 4],
        [3, 5, 5],
        [3, 6, 3],
        [4, 1, 1],
        [4, 1, 2],
        [4, 1, 3],
        [4, 1, 4],
        [4, 1, 5],
        [4, 2, 1],
        [4, 2, 2],
        [4, 2, 3],
        [4, 2, 4],
        [4, 2, 5],
        [4, 3, 1],
        [4, 3, 2],
        [4, 3, 3],
        [4, 3, 4],
        [4, 3, 5],
        [4, 4, 1],
        [4, 4, 2],
        [4, 4, 3],
        [4, 4, 4],
        [4, 4, 5],
        [4, 5, 1],
        [4, 5, 2],
        [4, 5, 3],
        [4, 5, 4],
        [4, 5, 5],
        [5, 1, 2],
        [5, 1, 3],
        [5, 1, 4],
        [5, 2, 1],
        [5, 2, 2],
        [5, 2, 3],
        [5, 2, 4],
        [5, 2, 5],
        [5, 3, 1],
        [5, 3, 2],
        [5, 3, 3],
        [5, 3, 4],
        [5, 3, 5],
        [5, 4, 1],
        [5, 4, 2],
        [5, 4, 3],
        [5, 4, 4],
        [5, 4, 5],
        [5, 5, 2],
        [5, 5, 3],
        [5, 5, 4],
        [6, 3, 3],
    ]
    return torch.tensor(out, device=device).T

def get_cached_disk_coords(device: str) -> Tensor:
    out = [
        [0, 7, 0],
        [1, 4, 0],
        [1, 5, 0],
        [1, 6, 0],
        [1, 7, 0],
        [1, 8, 0],
        [1, 9, 0],
        [1, 10, 0],
        [2, 3, 0],
        [2, 4, 0],
        [2, 5, 0],
        [2, 6, 0],
        [2, 7, 0],
        [2, 8, 0],
        [2, 9, 0],
        [2, 10, 0],
        [2, 11, 0],
        [3, 2, 0],
        [3, 3, 0],
        [3, 4, 0],
        [3, 5, 0],
        [3, 6, 0],
        [3, 7, 0],
        [3, 8, 0],
        [3, 9, 0],
        [3, 10, 0],
        [3, 11, 0],
        [3, 12, 0],
        [4, 1, 0],
        [4, 2, 0],
        [4, 3, 0],
        [4, 4, 0],
        [4, 5, 0],
        [4, 6, 0],
        [4, 7, 0],
        [4, 8, 0],
        [4, 9, 0],
        [4, 10, 0],
        [4, 11, 0],
        [4, 12, 0],
        [4, 13, 0],
        [5, 1, 0],
        [5, 2, 0],
        [5, 3, 0],
        [5, 4, 0],
        [5, 5, 0],
        [5, 6, 0],
        [5, 7, 0],
        [5, 8, 0],
        [5, 9, 0],
        [5, 10, 0],
        [5, 11, 0],
        [5, 12, 0],
        [5, 13, 0],
        [6, 1, 0],
        [6, 2, 0],
        [6, 3, 0],
        [6, 4, 0],
        [6, 5, 0],
        [6, 6, 0],
        [6, 7, 0],
        [6, 8, 0],
        [6, 9, 0],
        [6, 10, 0],
        [6, 11, 0],
        [6, 12, 0],
        [6, 13, 0],
        [7, 0, 0],
        [7, 1, 0],
        [7, 2, 0],
        [7, 3, 0],
        [7, 4, 0],
        [7, 5, 0],
        [7, 6, 0],
        [7, 7, 0],
        [7, 8, 0],
        [7, 9, 0],
        [7, 10, 0],
        [7, 11, 0],
        [7, 12, 0],
        [7, 13, 0],
        [7, 14, 0],
        [8, 1, 0],
        [8, 2, 0],
        [8, 3, 0],
        [8, 4, 0],
        [8, 5, 0],
        [8, 6, 0],
        [8, 7, 0],
        [8, 8, 0],
        [8, 9, 0],
        [8, 10, 0],
        [8, 11, 0],
        [8, 12, 0],
        [8, 13, 0],
        [9, 1, 0],
        [9, 2, 0],
        [9, 3, 0],
        [9, 4, 0],
        [9, 5, 0],
        [9, 6, 0],
        [9, 7, 0],
        [9, 8, 0],
        [9, 9, 0],
        [9, 10, 0],
        [9, 11, 0],
        [9, 12, 0],
        [9, 13, 0],
        [10, 1, 0],
        [10, 2, 0],
        [10, 3, 0],
        [10, 4, 0],
        [10, 5, 0],
        [10, 6, 0],
        [10, 7, 0],
        [10, 8, 0],
        [10, 9, 0],
        [10, 10, 0],
        [10, 11, 0],
        [10, 12, 0],
        [10, 13, 0],
        [11, 2, 0],
        [11, 3, 0],
        [11, 4, 0],
        [11, 5, 0],
        [11, 6, 0],
        [11, 7, 0],
        [11, 8, 0],
        [11, 9, 0],
        [11, 10, 0],
        [11, 11, 0],
        [11, 12, 0],
        [12, 3, 0],
        [12, 4, 0],
        [12, 5, 0],
        [12, 6, 0],
        [12, 7, 0],
        [12, 8, 0],
        [12, 9, 0],
        [12, 10, 0],
        [12, 11, 0],
        [13, 4, 0],
        [13, 5, 0],
        [13, 6, 0],
        [13, 7, 0],
        [13, 8, 0],
        [13, 9, 0],
        [13, 10, 0],
        [14, 7, 0],
        [0, 3, 1],
        [1, 1, 1],
        [1, 2, 1],
        [1, 3, 1],
        [1, 4, 1],
        [1, 5, 1],
        [2, 1, 1],
        [2, 2, 1],
        [2, 3, 1],
        [2, 4, 1],
        [2, 5, 1],
        [3, 0, 1],
        [3, 1, 1],
        [3, 2, 1],
        [3, 3, 1],
        [3, 4, 1],
        [3, 5, 1],
        [3, 6, 1],
        [4, 1, 1],
        [4, 2, 1],
        [4, 3, 1],
        [4, 4, 1],
        [4, 5, 1],
        [5, 1, 1],
        [5, 2, 1],
        [5, 3, 1],
        [5, 4, 1],
        [5, 5, 1],
        [6, 3, 1],
        [0, 3, -1],
        [1, 1, -1],
        [1, 2, -1],
        [1, 3, -1],
        [1, 4, -1],
        [1, 5, -1],
        [2, 1, -1],
        [2, 2, -1],
        [2, 3, -1],
        [2, 4, -1],
        [2, 5, -1],
        [3, 0, -1],
        [3, 1, -1],
        [3, 2, -1],
        [3, 3, -1],
        [3, 4, -1],
        [3, 5, -1],
        [3, 6, -1],
        [4, 1, -1],
        [4, 2, -1],
        [4, 3, -1],
        [4, 4, -1],
        [4, 5, -1],
        [5, 1, -1],
        [5, 2, -1],
        [5, 3, -1],
        [5, 4, -1],
        [5, 5, -1],
        [6, 3, -1],
    ]

    out = torch.tensor(out, device=device).T
    out[0,:] -= 8
    out[1,:] -= 8
    return out


"""
For code used in distributed training.
"""
from typing import Tuple

import torch
import torch.distributed as dist
import torch.multiprocessing as mp
from torch import Tensor


def set_sharing_strategy(new_strategy=None):
    """
    https://pytorch.org/docs/stable/multiprocessing.html
    https://discuss.pytorch.org/t/how-does-one-setp-up-the-set-sharing-strategy-strategy-for-multiprocessing/113302
    https://stackoverflow.com/questions/66426199/how-does-one-setup-the-set-sharing-strategy-strategy-for-multiprocessing-in-pyto
    """
    from sys import platform

    if new_strategy is not None:
        mp.set_sharing_strategy(new_strategy=new_strategy)
    else:
        if platform == "darwin":  # OS X
            # only sharing strategy available at OS X
            mp.set_sharing_strategy("file_system")
        else:
            # ulimit -n 32767 or ulimit -n unlimited (perhaps later do try catch to execute this increase fd limit)
            mp.set_sharing_strategy("file_descriptor")


def use_file_system_sharing_strategy():
    """
    when to many file descriptor error happens

    https://discuss.pytorch.org/t/how-does-one-setp-up-the-set-sharing-strategy-strategy-for-multiprocessing/113302
    """
    import torch.multiprocessing

    torch.multiprocessing.set_sharing_strategy("file_system")


def find_free_port():
    """https://stackoverflow.com/questions/1365265/on-localhost-how-do-i-pick-a-free-port-number"""
    import socket
    from contextlib import closing

    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return str(s.getsockname()[1])


def setup_process(rank, world_size, port, backend="gloo"):
    """
    Initialize the distributed environment (for each process).

    gloo: is a collective communications library (https://github.com/facebookincubator/gloo). My understanding is that
    it's a library/API for process to communicate/coordinate with each other/master. It's a backend library.

    export NCCL_SOCKET_IFNAME=eth0
    export NCCL_IB_DISABLE=1

    https://stackoverflow.com/questions/61075390/about-pytorch-nccl-error-unhandled-system-error-nccl-version-2-4-8

    https://pytorch.org/docs/stable/distributed.html#common-environment-variables
    """
    import torch.distributed as dist
    import os
    import torch

    if rank != -1:  # -1 rank indicates serial code
        print(f"setting up rank={rank} (with world_size={world_size})")
        # MASTER_ADDR = 'localhost'
        MASTER_ADDR = "127.0.0.1"
        # set up the master's ip address so this child process can coordinate
        os.environ["MASTER_ADDR"] = MASTER_ADDR
        print(f"{MASTER_ADDR=}")
        os.environ["MASTER_PORT"] = port
        print(f"{port=}")

        # - use NCCL if you are using gpus: https://pytorch.org/tutorials/intermediate/dist_tuto.html#communication-backends
        if torch.cuda.is_available():
            # unsure if this is really needed
            # os.environ['NCCL_SOCKET_IFNAME'] = 'eth0'
            # os.environ['NCCL_IB_DISABLE'] = '1'
            backend = "nccl"
        print(f"{backend=}")
        # Initializes the default distributed process group, and this will also initialize the distributed package.
        dist.init_process_group(backend, rank=rank, world_size=world_size)
        # dist.init_process_group(backend, rank=rank, world_size=world_size)
        # dist.init_process_group(backend='nccl', init_method='env://', world_size=world_size, rank=rank)
        print(f"--> done setting up rank={rank}")


def cleanup(rank):
    """Destroy a given process group, and deinitialize the distributed package"""
    # only destroy the process distributed group if the code is not running serially
    if rank != -1:  # -1 rank indicates serial code
        dist.destroy_process_group()


def get_batch(batch: Tuple[Tensor, Tensor], rank) -> Tuple[Tensor, Tensor]:
    x, y = batch
    if torch.cuda.is_available():
        x, y = x.to(rank), y.to(rank)
    else:
        # I don't think this is needed...
        # x, y = x.share_memory_(), y.share_memory_()
        pass
    return x, y


def test_setup():
    print("test_setup")
    port = find_free_port()
    world_size = 2
    mp.spawn(setup_process, args=(world_size, port), nprocs=2)
    print("successful test_setup!")


if __name__ == "__main__":
    test_setup()
